from sensiml.method_calls.selectorcall import SelectorCall
from sensiml.datamanager.pipeline import PipelineStep


class SelectorCallSet(PipelineStep):
    """The base class for a collection of selector calls"""

    def __init__(self, name=""):
        super(SelectorCallSet, self).__init__(name=name, step_type="SelectorCallSet")
        self._selectors = []
        self._input_data = ""
        self._label_column = None
        self._number_of_features = 10
        self._feature_table = ""
        self._cost_function = ""
        self._remove_columns = []
        self._passthrough_columns = None
        self._refinement = {}

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def input_data(self):
        return self._input_data

    @input_data.setter
    def input_data(self, value):
        self._input_data = value

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._label_column = value

    @property
    def number_of_features(self):
        return self._number_of_features

    @number_of_features.setter
    def number_of_features(self, value):
        self._number_of_features = value

    @property
    def feature_table(self):
        return self._feature_table

    @feature_table.setter
    def feature_table(self, value):
        self._feature_table = value

    @property
    def cost_function(self):
        return self._cost_function

    @cost_function.setter
    def cost_function(self, value):
        self._cost_function = value

    @property
    def refinement(self):
        return self._refinement

    @refinement.setter
    def refinement(self, value):
        self._refinement = value

    @property
    def selectors(self):
        return self._selectors

    def add_selector_call(self, *selectors):
        """Adds one or more selector calls to the collection.

        Args:
            selectors (SelectorCall or list[SelectorCall]): object(s) to append
        """
        for selector in selectors:
            self._selectors.append(selector)

    def remove_selector_call(self, *selectors):
        """Removes one or more selector call from the collection.

        Args:
            selectors (SelectorCall or list[SelectorCall]): object(s) to remove
        """
        for selector in selectors:
            self._selectors = [f for f in self._selectors if f != selector]
        # Note:  We should probably re-populate inputs and outputs here

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def passthrough_columns(self):
        return self._passthrough_columns

    @passthrough_columns.setter
    def passthrough_columns(self, value):
        self._passthrough_columns = value

    @property
    def remove_columns(self):
        return self._remove_columns

    @remove_columns.setter
    def remove_columns(self, value):
        self._remove_columns = value

    def _to_list(self):
        gencalls = []
        for item in self._selectors:
            gencalls.append(item._to_dict())
        return gencalls

    def _to_dict(self):
        selcalls_set = []
        set_dict = {}
        for item in self._selectors:
            selcalls_set.append(item._to_dict())
        set_dict["type"] = "selectorset"
        set_dict["name"] = getattr(self, "_name")
        set_dict["set"] = selcalls_set
        set_dict["inputs"] = {}
        set_dict["inputs"]["remove_columns"] = getattr(self, "_remove_columns")
        set_dict["inputs"]["passthrough_columns"] = getattr(
            self, "_passthrough_columns"
        )
        set_dict["inputs"]["input_data"] = getattr(self, "_input_data")
        set_dict["inputs"]["label_column"] = getattr(self, "_label_column")
        set_dict["inputs"]["number_of_features"] = getattr(self, "_number_of_features")
        set_dict["inputs"]["feature_table"] = getattr(self, "_feature_table")
        if self._cost_function == "":
            self._cost_function = "sum"
        set_dict["inputs"]["cost_function"] = getattr(self, "_cost_function")
        set_dict["outputs"] = self._outputs
        set_dict["refinement"] = getattr(self, "_refinement")
        return set_dict



import json, yaml
import serial
from decimal import *
from functools import partial, reduce
from datetime import timedelta


class ProfileParser(object):
    def __init__(self, json_path, port, num_msgs=100, baud=115200, debug=False):
        self._port = port
        self.baud = baud
        self._is_running = False
        self._feature_times = list()
        self._feature_average_time = list()
        self._feature_cycles = list()
        self._feature_average_cycles = list()
        self._classifier_times = list()
        self._classifier_average_time = list()
        self._classifier_cycles = list()
        self._classifier_average_cycles = list()
        self._is_running = False
        self._combined_metrics = dict()
        self._max_msgs = num_msgs
        self._debug = debug

        if json_path:
            with open(json_path, "r") as f:
                self._model_desc = json.load(f)

            for model_index, model_name in self._model_desc["ModelIndexes"].items():
                mi = int(model_index)
                model_type = self._model_desc["ModelDescriptions"][mi]["ModelType"]
                self._feature_times.insert(mi, dict())
                self._feature_average_time.insert(mi, dict())
                self._feature_cycles.insert(mi, dict())
                self._feature_average_cycles.insert(mi, dict())
                self._classifier_times.insert(mi, dict())
                self._classifier_average_time.insert(mi, dict())
                self._classifier_cycles.insert(mi, dict())
                self._classifier_average_cycles.insert(mi, dict())
                for feature in self._model_desc["ModelDescriptions"][mi][
                    "FeatureFunctions"
                ]:
                    self._feature_times[mi][feature] = list()
                    self._feature_average_time[mi][feature] = 0
                    self._feature_cycles[mi][feature] = list()
                    self._feature_average_cycles[mi][feature] = 0
                self._classifier_times[mi][model_type] = list()
                self._classifier_average_time[mi][model_type] = 0
                self._classifier_cycles[mi][model_type] = list()
                self._classifier_average_cycles[mi][model_type] = 0

        else:
            self._model_desc = None

    def run(self):
        msgs_rx = 0
        self._is_running = True
        self._ser = serial.Serial(port=self._port, baudrate=self.baud)
        while msgs_rx < self._max_msgs:
            try:
                line = self._ser.readline()
                decoded = json.loads(line)
                self._parse_message(decoded)
                msgs_rx += 1
            except Exception as e:
                if self._debug:
                    print(e)
                    print(line)
                continue
        self._ser.close()
        self._combined_metrics.update(
            {
                "FeatureTimePerInference": self.feature_average_time,
                "FeatureCyclesPerInference": self.feature_average_cycles,
                "ClassifierAverageTime": self.classifier_average_time,
                "ClassifierAverageCycles": self.classifier_average_cycles,
                "RunCount": msgs_rx,
            }
        )
        print("Reader closed")

    def stop(self):
        self._is_running = False

    @property
    def combined_metrics(self):
        return self._combined_metrics

    @property
    def is_running(self):
        return self._is_running

    @property
    def feature_times(self):
        return self._feature_times

    @property
    def feature_average_time(self):
        return self._feature_average_time

    @property
    def classifier_times(self):
        return self._classifier_times

    @property
    def classifier_average_time(self):
        return self._classifier_average_time

    @property
    def feature_cycles(self):
        return self._feature_cycles

    @property
    def feature_average_cycles(self):
        return self._feature_average_cycles

    @property
    def classifier_cycles(self):
        return self._classifier_cycles

    @property
    def classifier_average_cycles(self):
        return self._classifier_average_cycles

    def _average(self, lst):
        return reduce(lambda a, b: a + b, lst) / len(lst)

    def _format_result(self, result):
        seconds = int(result)
        microseconds = (result * 1000000) % 1000000
        nanoseconds = (result * 1000000000) % 1000000000
        output = Decimal((microseconds / 1000000) + nanoseconds / 1000000000)
        output = output.quantize(Decimal("0.0000000001"))
        return output

    def _parse_message(self, data):
        model_number = data.get("ModelNumber", 0)
        string_type = data.get("Type", None)
        model_type = self._model_desc["ModelDescriptions"][model_number]["ModelType"]
        feature_index = 0
        if string_type == "Cycles":
            for k in self._feature_times[model_number].keys():
                cycles_per_sample = data["FeatureCycles"][feature_index]
                self._feature_cycles[model_number][k].append(cycles_per_sample)
                self._feature_average_cycles[model_number][k] = round(
                    self._average(self._feature_cycles[model_number][k])
                )
                self._classifier_cycles[int(model_number)][model_type].append(
                    data["ClassifierCycles"]
                )
                feature_index += 1
            self._classifier_average_cycles[int(model_number)][model_type] = round(
                self._average(self._classifier_cycles[model_number][model_type])
            )
        elif string_type == "Times":
            for k in self._feature_times[model_number].keys():
                time_per_sample = data["FeatureTimes"][feature_index]
                self._feature_times[model_number][k].append(time_per_sample)
                self._feature_average_time[model_number][k] = "{0:0.9f}".format(
                    self._format_result(
                        self._average(self._feature_times[model_number][k])
                    )
                )
                self._classifier_times[int(model_number)][model_type].append(
                    data["ClassifierTime"]
                )

                feature_index += 1
            self._classifier_average_time[int(model_number)][
                model_type
            ] = "{0:0.9f}".format(
                self._format_result(
                    self._average(self._classifier_times[model_number][model_type])
                )
            )

    def to_json(self):
        return json.dumps(self._combined_metrics)

    def to_yaml(self):
        return yaml.dump(self._combined_metrics)



import os
import sqlite3
import uuid
import wave
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame

from sensiml.dclproj.datasegments import (
    DataSegment,
    DataSegments,
    segment_list_to_datasegments,
)
from sensiml.dclproj.utils import mfcc, plot_segments_labels
from sensiml.dclproj.vizualization import plot_segments


class DCLProject:
    """
    The DCLProject class provides read-only access to the .dclproj file that is associated with a Data Capture Lab project. The
    are a number of helper functions for visualizations and queries.


    dclproj_path = '<PATH-To-File.dclproj>'

    dcl = DCLProject(path=dclproj_path)


    """

    def __init__(self, path: Optional[str] = None):
        self._path = path
        self._conn = None
        self._tables = None
        self._verbose = False

        if self._path:
            self.create_connection(self._path)

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool):
        if isinstance(value, bool):
            self._verbose = value
        else:
            print("verbose must be either True or False.")

    @property
    def data_dir(self) -> str:

        if self._path is None:
            raise Exception("Path is not set!")

        return os.path.join(self._path, "data")

    def _set_table_list(self):
        cursorObj = self._conn.cursor()

        cursorObj.execute('SELECT name from sqlite_master where type= "table"')

        self._tables = [x[0] for x in cursorObj.fetchall()]

    def create_connection(self, db_file: str) -> None:
        """create a database connection to the SQLite database specified by db_file

        :param db_file: database file
        :return: None
        """
        self._path = os.path.dirname(db_file)
        if not os.path.exists(db_file):
            print("database file not found at {}".format(db_file))
        conn = None
        try:
            conn = sqlite3.connect(db_file)
        except Exception as e:
            print(e)
            raise e

        self._conn = conn
        self._set_table_list()

    def _execute_query(self, query: str, header: str = None) -> DataFrame:

        cur = self._conn.cursor()

        if self.verbose:
            print(query)

        cur.execute(query)

        if header is None:
            header = [x[0] for x in cur.description]

        rows = cur.fetchall()

        results = []
        for row in rows:
            results.append(row)

        df = pd.DataFrame(results, columns=header)

        if "uuid" in header:
            df["uuid"] = df["uuid"].apply(
                lambda x: uuid.UUID(bytes_le=x) if x else None
            )

        if "last_modified" in header:
            df["last_modified"] = pd.to_datetime(df["last_modified"])

        if "created_at" in header:
            df["created_at"] = pd.to_datetime(df["created_at"])

        if "local_status" in header:
            local_status = {
                0: "To Add",
                1: "To Update",
                "Synced": "Synced",
                2: "To Delete",
                None: "Synced",
            }
            df["local_status"] = (
                df["local_status"].fillna("Synced").apply(lambda x: local_status[x])
            )

        return df

    def _list_table_raw(self, tablename: str, filter_deleted: bool = True) -> DataFrame:

        if tablename not in self._tables:
            print("Table is not part of the database.")
            return None

        query = "SELECT * FROM {tablename} ".format(tablename=tablename)

        if filter_deleted:
            query += f"WHERE ({tablename}.local_status != 2 OR {tablename}.local_status IS NULL)"

        return self._execute_query(query)

    def _list_table(
        self,
        tablename: str,
        fields: List[str],
        fk_fields: Optional[Dict] = None,
        query_filter: Optional[str] = None,
        header=None,
        filter_deleted: bool = True,
    ):

        if tablename not in self._tables:
            print("Table is not part of the database.")
            return None

        select_fields = ", ".join(
            ["{}.{}".format(tablename, field) for field in fields]
        )

        if fk_fields:
            select_fields += ", " + ", ".join(
                [
                    "{}.name".format(fk_table)
                    if fk_table != "LabelValue"
                    else "{}.value".format(fk_table)
                    for fk_table in fk_fields.keys()
                ]
            )

        query_select = "SELECT {select_fields} FROM {tablename} ".format(
            select_fields=select_fields, tablename=tablename
        )

        query_join = ""

        if fk_fields:
            join_fields = []
            for fk_table, fk_field in fk_fields.items():
                join_fields.append(
                    "LEFT JOIN {fk_table} ON {fk_table}.id = {tablename}.{fk_field} ".format(
                        fk_table=fk_table, tablename=tablename, fk_field=fk_field
                    )
                )
            query_join += " ".join(join_fields)

        if query_filter and filter_deleted:
            query_where = " ".join(
                [
                    query_filter,
                    f"AND ({tablename}.local_status != 2  OR {tablename}.local_status IS NULL)",
                ]
            )
        elif filter_deleted:
            query_where = f" WHERE ({tablename}.local_status != 2  OR {tablename}.local_status IS NULL)"
        else:
            query_where = " "

        if fk_fields and not header:
            header = fields + list(fk_fields.values())

        return self._execute_query(query_select + query_join + query_where, header)

    def _list_captures_metadata(self) -> DataFrame:
        """returns a dataframe containing the capture name and the associated metadata"""

        fields = ["uuid"]

        fk_fields = OrderedDict(
            [("Capture", "capture"), ("LabelValue", "label_value"), ("Label", "label")]
        )

        df = self._list_table("CaptureMetadataValue", fields, fk_fields=fk_fields)

        M = []
        for capture_name in df.capture.unique():
            caps_dict = df[df.capture == capture_name][
                ["label", "label_value"]
            ].to_dict(orient="records")
            tmp = {"capture": capture_name}
            for item in caps_dict:
                tmp[item["label"]] = item["label_value"]
            M.append(tmp)

        return DataFrame(M).fillna("")

    def list_captures(self, include_metadata: bool = True) -> DataFrame:
        """List the captures in the DCLI project file

        Args:
            include_metadata (bool, optional): If True return the associated metadata information with each capture. Defaults to True.

        Returns:
            DataFrame: DataFrame containing the capture information
        """

        fields = [
            "uuid",
            "name",
            "file_size",
            "number_samples",
            "set_sample_rate",
            "created_at",
            "local_status",
            "last_modified",
        ]

        fk_fields = OrderedDict([("CaptureConfiguration", "capture_configuration")])

        df = self._list_table("Capture", fields=fields, fk_fields=fk_fields)

        if not include_metadata or self._list_captures_metadata().shape[0] == 0:
            return df

        return df.merge(
            self._list_captures_metadata(),
            left_on="name",
            right_on="capture",
            how="left",
        ).drop("capture", axis=1)

    def list_sessions(self) -> DataFrame:
        fields = [
            "id",
            "name",
            "parameters",
            "custom",
            "preprocess",
            "created_at",
            "local_status",
            "last_modified",
        ]

        fk_fields = OrderedDict([])

        return self._list_table("Segmenter", fields, fk_fields)

    def list_capture_segments(
        self,
        captures: Optional[List] = None,
        sessions: Optional[List] = None,
        include_ids: bool = False,
    ) -> DataFrame:
        """Returns a DataFrame of segment information that are in the captures and sessions specified

        Args:
            captures (Optional[List], optional): A list of captures to return segments about. Defaults to None.
            sessions (Optional[List], optional): A list of sessions to return segments for. Defaults to None.

        Returns:
            DataFrame:
        """

        if captures is not None and isinstance(captures, str):
            captures = [captures]
        if sessions is not None and isinstance(sessions, str):
            sessions = [sessions]

        query_filter = ""

        def create_select(select):
            return ", ".join(['"{}"'.format(x) for x in select])

        if captures and sessions:
            query_filter = "WHERE Capture.name in ({}) ".format(
                create_select(captures)
            ) + "AND Segmenter.name in ({0}) ".format(create_select(sessions))

        elif captures:
            query_filter = "WHERE Capture.name in ({}) ".format(create_select(captures))
        elif sessions:
            query_filter += "WHERE Segmenter.name in ({0}) ".format(
                create_select(sessions)
            )

        fields = [
            "uuid",
            "capture_sample_sequence_start",
            "capture_sample_sequence_end",
            "created_at",
            "last_modified",
        ]

        fk_fields = OrderedDict(
            [
                ("Segmenter", "segmenter"),
                ("Capture", "capture"),
                ("LabelValue", "label_value"),
                ("Label", "label"),
            ]
        )

        if include_ids:
            fields += ["segmenter", "capture", "label", "label_value"]
            header = [
                "uuid",
                "capture_sample_sequence_start",
                "capture_sample_sequence_end",
                "created_at",
                "last_modified",
                "segmenter_id",
                "capture_id",
                "label_id",
                "label_value_id",
                "segmenter",
                "capture",
                "label_value",
                "label",
            ]
        else:
            header = None

        return (
            self._list_table(
                "CaptureLabelValue", fields, fk_fields, query_filter, header=header
            )
            .reset_index(drop=True)
            .sort_values(by=["capture", "segmenter", "capture_sample_sequence_start"])
        )

    def get_capture_stats(self, capture: str) -> Dict:
        """Returns information about the capture

        Args:
            capture (str): name of capture

        Returns:
            Dict: information about the capture
        """
        capture_list = self.list_captures()

        if capture not in capture_list.name.values:
            raise Exception("Capture not found!")

        return capture_list[capture_list.name == capture].to_dict(orient="records")[0]

    def get_capture_metadata(
        self, capture: str, include_ids: bool = False
    ) -> DataFrame:
        """Gets the metadata for the specified capture

        Args:
            capture (str): name of the capture file

        Returns:
            DataFrame: DataFrame containing the metadata associated with this capture
        """

        fields = ["uuid"]

        fk_fields = OrderedDict(
            [("Capture", "capture"), ("LabelValue", "label_value"), ("Label", "label")]
        )

        if include_ids:
            fields += ["capture", "label_value", "label"]
            header = header = [
                "uuid",
                "capture_id",
                "label_value_id",
                "label_id",
                "capture",
                "label_value",
                "label",
            ]
        else:
            header = None

        df = self._list_table(
            "CaptureMetadataValue", fields, fk_fields=fk_fields, header=header
        )

        return df[df["capture"] == capture]

    def get_capture_segments(
        self, captures: Optional[List] = None, sessions: Optional[List] = None
    ):
        """
        Returns a DataSegment object of the specified capture and session


        Args:
            capture_name (str): name of capture
            session (str):  name of session where the labels are
        """
        if captures is not None and isinstance(captures, str):
            captures = [captures]
        if sessions is not None and isinstance(sessions, str):
            sessions = [sessions]

        labels = self.list_capture_segments(captures=captures, sessions=sessions)

        data_segments = DataSegments()

        cached_df = None
        capture_index = 4
        for index, label in enumerate(
            labels[
                [
                    "capture_sample_sequence_start",
                    "capture_sample_sequence_end",
                    "uuid",
                    "label_value",
                    "capture",
                    "segmenter",
                ]
            ].values
        ):
            # TODO: read wave file if file is .wav#
            if cached_df != label[capture_index]:
                tmp_df = self.get_capture(label[capture_index])
                cached_df = label[capture_index]

            data_segments.append(
                DataSegment(
                    segment_id=index,
                    capture_sample_sequence_start=label[0],
                    capture_sample_sequence_end=label[1],
                    data=tmp_df.loc[label[0] : label[1]].values.T,
                    columns=tmp_df.columns.to_list(),
                    session=label[5],
                    label_value=label[3],
                    uuid=label[2],
                    capture=label[4],
                )
            )

        return data_segments

    def get_capture(self, capture_name: str) -> DataFrame:
        """
        Returns the capture as a DataFrame

        Args:
            capture_name (str): name of capture
        """

        # TODO: read wave file if file is .wav
        ext = capture_name.split(".")[-1]

        if not os.path.exists(os.path.join(self.data_dir, capture_name)):
            raise Exception(
                "File {} does not exist!".format(
                    os.path.join(self.data_dir, capture_name)
                )
            )

        if ext == "csv":
            tmp_df = pd.read_csv(
                os.path.join(self.data_dir, capture_name), index_col="sequence"
            )
        elif ext == "wav":
            with wave.open(
                os.path.join(self.data_dir, capture_name), "rb"
            ) as wave_reader:
                waveFrames = wave_reader.readframes(wave_reader.getnframes())
                waveData = np.fromstring(waveFrames, dtype=np.int16).reshape(
                    (-1, wave_reader.getnchannels())
                )
                columns = [
                    "channel_{}".format(i) for i in range(wave_reader.getnchannels())
                ]
                tmp_df = DataFrame(waveData, columns=columns).rename_axis(
                    index="sequence"
                )

        return tmp_df

    def get_captures(self, capture_names: Optional[List[str]] = None) -> DataFrame:
        """
        Returns the capture as a list of dataframes

        Args:
            capture_names (List[str]): name of captures to return, if None returns all captures
        """
        if capture_names is None:
            capture_names = os.listdir(self.data_dir)

        return {
            capture_name: self.get_capture(capture_name)
            for capture_name in capture_names
        }

    def plot_segment_labels(
        self,
        capture_names: List[str],
        session: str,
        columns: Optional[List[str]] = None,
        transforms: Optional[List[Dict]] = None,
        figsize: Tuple = (30, 8),
        ylim: Optional[Tuple] = None,
        xlim: Optional[Tuple] = None,
    ) -> None:
        """
        Creates a plot of the labels and raw signal data for a session and one or more capture files.

        Args:
            capture_names (List[str]): List if captures to plot
            session (str):  name of session to pull the labels from
            columns (list): a list of columns to plot from the data
             label the value is the label it was should be renamed to
            transforms: additional fucntions to apply prior to plotting the data


        Examples:
            >>> def sum_columns(x):
            >>>     return x[['Column1', 'Column2']].sum()

            >>> func = {'name':'sum','func':sum_columns, 'type':'transform'}

            >>> dcl.plot_segment_labels(captures, session_name, columns=['sum'], transforms=[func])

            >>> # example of multiple transforms executed in order

            >>> def sum_columns(x):
            >>>     return x[['Column1', 'Column2']].sum()

            >>> def remove_offset(x):
            >>>    x['Column1']-=100
            >>>    x['Column2']-=200
            >>>    return x

            >>> func2 = {'name':'filter','func':remove_offset, "type":'filter', "columns":['Column1', 'Column2']}
            >>> func1 = {'name':'sum','func':sum_columns, "type":'transform'}

            >>> dcl.plot_segment_labels(captures, session_name, columns=['Column1', 'Column2', 'sum'], transforms=[func1, func2])
        """

        sessions = self.list_sessions()
        session_id = sessions[sessions.name == session].id.values[0]
        clv = self._list_table_raw("CaptureLabelValue")
        label_value_ids = clv[clv.segmenter == session_id].label_value.unique()
        lv = self._list_table_raw("LabelValue")
        labels = sorted(lv[lv.id.isin(label_value_ids)].value.values.tolist())
        class_map = {label: index for index, label in enumerate(labels)}

        M = []
        current_start = 0

        plt.figure(figsize=figsize)
        currentAxis = plt.gca()

        if isinstance(capture_names, str):
            capture_names = [capture_names]

        for capture_name in capture_names:

            segments = self.list_capture_segments(
                captures=[capture_name], sessions=[session]
            )

            if xlim:
                segments = segments[
                    (segments.capture_sample_sequence_start < xlim[1])
                    & (segments.capture_sample_sequence_end > xlim[0])
                ]

            capture = self.get_capture(capture_name)
            M.append(capture)

            plot_segments(
                segments,
                capture,
                capture_name=capture_name,
                labels=list(class_map.keys()),
                current_start=current_start,
                ylim=ylim,
                figsize=figsize,
                currentAxis=currentAxis,
            )

            current_start += capture.shape[0]

        capture = pd.concat(M)

        if transforms is not None:
            for func in transforms:
                if func["type"] == "transform":
                    capture[func["name"]] = capture.apply(
                        lambda x: func["func"](x), axis=func.get("axis", 1)
                    )

                if func["type"] == "filter":
                    capture[func["columns"]] = capture.apply(
                        lambda x: func["func"](x), axis=func.get("axis", 1)
                    )

        if columns is None:
            columns = capture.columns

        capture[columns].reset_index(drop=True).plot(
            figsize=figsize,
            ax=currentAxis,
            title="session: '{session}'".format(capture=capture_name, session=session),
            ylim=ylim,
            xlim=xlim,
        )

        return capture.reset_index(drop=True)

    def plot_frequency(
        self,
        capture_name: str,
        channel: str,
        sample_freq: int = 16000,
        figsize: Tuple = (30, 4),
        session: Optional[str] = None,
    ):

        nrows = 3
        data = self.get_capture(capture_name)

        fig, axes = plt.subplots(
            nrows=nrows, ncols=1, figsize=(figsize[0], figsize[1] * nrows)
        )
        fig.tight_layout()

        data.plot(ax=axes[0], title=capture_name, xlim=(0, data.shape[0]), xlabel="")

        _ = axes[1].specgram(data[channel], NFFT=512)
        axes[1].set_title("Sepctrogram")

        axes[2].imshow(
            mfcc(data, channel, sample_freq=sample_freq, plot_all=False), aspect="auto"
        )
        axes[2].set_title("MFCC")

        if session:
            segments = self.list_capture_segments(
                captures=[capture_name], sessions=[session]
            )

            plot_segments_labels(
                segments,
                axes=axes[0],
                y_label=(
                    data[channel].min(),
                    int(data[channel].max()) - int(data[channel].min()),
                ),
            )

        plt.subplots_adjust(hspace=0.2)

        return axes



import json
from pandas import DataFrame, Series

from sensiml.datamanager.clientplatformdescription import ClientPlatformDescription
import sensiml.base.utility as utility


class ClientPlatformDescriptions:
    """Base class for a collection of functions"""

    def __init__(self, connection):
        self._connection = connection
        self.build_platform_descriptions()

    def __getitem__(self, index):
        if len(self.platform_list) == 0:
            self.build_platform_descriptions()

        df = self()
        for p in self.platform_list:
            if index not in df.index:
                continue
            if df.loc[index].equals(p().loc[0]):
                return p
        return None

    def refresh(self):
        self.build_platform_descriptions()

    def build_platform_descriptions(self):
        """Populates the platform_list property from the server."""
        self.platform_list = []
        self.platform_dict = {}

        platforms = self.get_platforms()
        for platform in platforms:
            self.platform_dict["{}".format(platform.name)] = platform
            self.platform_list.append(platform)

    def get_platform_by_name(self, name):
        if len(self.platform_list) == 0:
            self.build_platform_descriptions()
        return self.platform_dict.get(name, None)

    def get_platform_by_id(self, id):
        if len(self.platform_list) == 0:
            self.build_platform_descriptions()

        return next((x for x in self.platform_list if x.uuid == id), None)

    def get_platform_by_uuid(self, uuid):
        return self.get_platform_by_id(uuid)

    def _new_platform_description_from_dict(self, data):
        """Creates and populates a platform from a set of properties.

        Args:
            data (dict): contains properties of a platform

        Returns:
            platform
        """
        platform = ClientPlatformDescription(self._connection)
        platform.initialize_from_dict(data)
        return platform

    def get_platforms(self, function_type=""):
        """Gets all functions as function objects.

        Args:
            function_type (optional[str]): type of function to retrieve

        Returns:
            list of functions
        """
        url = "platforms/v2"
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        platformDescriptions = list()
        for platformdesc in response_data:
            platformDescriptions.append(
                self._new_platform_description_from_dict(platformdesc)
            )

        return platformDescriptions

    def list_processors(self):
        all_platforms = self.get_platforms()
        procs = {"name": {}}
        for plat in all_platforms:
            for processor in plat.processors:
                if processor.display_name not in procs["name"].keys():
                    procs["name"][processor.display_name] = processor.uuid
        return DataFrame(procs)

    def __call__(self):
        return self.__str__()

    def __str__(self):
        all_platforms = self.get_platforms()
        if len(all_platforms) < 0:
            return DataFrame()
        ret = DataFrame(p.__dict__() for p in all_platforms)
        # for plat in all_platforms:
        #     ret = ret.append(plat(), ignore_index=True)
        return ret.sort_values(by="Id").set_index("Id", drop=True)



from sensiml.method_calls.basemethodcall import BaseMethodCall


class ClassifierCall(BaseMethodCall):
    """The base class for a classifier call"""

    def __init__(self, name, function_type="classifier"):
        super(ClassifierCall, self).__init__(name=name, function_type=function_type)



from sensiml.datamanager.pipeline import PipelineStep


class DataFileCall(PipelineStep):
    """The base class for a featurefile call"""

    def __init__(self, name):
        super(DataFileCall, self).__init__(name=name, step_type="DataFileCall")
        self.name = name
        self._data_columns = None
        self._group_columns = None
        self._label_column = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, list):
            name = [name]
        self._name = name

    @property
    def data_columns(self):
        return self._data_columns

    @data_columns.setter
    def data_columns(self, value):
        self._data_columns = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._label_column = value

    @property
    def group_columns(self):
        return self._group_columns

    @group_columns.setter
    def group_columns(self, value):
        self._group_columns = value

    def _to_dict(self):
        capturefile_dict = {}
        capturefile_dict["name"] = self.name
        capturefile_dict["type"] = "datafile"
        capturefile_dict["data_columns"] = self.data_columns
        capturefile_dict["group_columns"] = self.group_columns
        capturefile_dict["label_column"] = self.label_column
        capturefile_dict["outputs"] = self.outputs
        return capturefile_dict



class FunctionCalls(object):
    """Collection of function calls"""

    def __init__(self):
        self._name = ""
        self._function_calls = []

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def function_calls(self):
        return self._function_calls

    def add_function_call(self, function_call):
        """Adds a function call to the collection.

        Args:
            function_call (FunctionCall): object to append
        """
        self._function_calls.append(function_call)

    def remove_function_call(self, function_call):
        """Removes a function call from the collection.

        Args:
            function_call (FunctionCall): object to remove
        """
        self._function_calls = [f for f in self._function_calls if f != function_call]

    def _to_list(self):
        fcalls = []
        for item in self.function_calls:
            fcalls.append(item._to_dict())
        return fcalls

        # fcalls['name'] = self.name
        # fcalls['type'] = 'transform'
        # query_dict['outputs'] = [self.outputs]



from sensiml.method_calls.basemethodcall import BaseMethodCall


class GeneratorCall(BaseMethodCall):
    """The base class for a generator call.

    Child classes have their own additional properties and overwrite this docstring."""

    def __init__(self, name, function_type="Feature Generator"):
        super(GeneratorCall, self).__init__(name=name, function_type=function_type)

    def _to_dict(self):
        prop_dict = {}
        if not hasattr(self, "_subtype"):
            prop_dict["function_name"] = self._name
        else:
            prop_dict["subtype"] = self._subtype

        prop_dict["inputs"] = {}
        for prop in self._public_properties():
            prop_dict["inputs"][prop] = getattr(self, prop)
        return prop_dict



try:
    from http.cookiejar import CookiePolicy
except ImportError:
    from cookielib import CookiePolicy

import errno
import json as JSON
import logging
import os
import re
import warnings
from datetime import datetime, timezone
from functools import partial
from getpass import getpass
from time import sleep
from uuid import uuid4

import requests
import requests_oauthlib
from appdirs import user_config_dir
from oauthlib.oauth2 import (
    InvalidGrantError,
    LegacyApplicationClient,
    MissingTokenError,
    TokenExpiredError,
)
from requests import Response
from requests_oauthlib import OAuth2, OAuth2Session
from six.moves import input

from .connection_config import ConnectionConfig

try:
    import configparser as ConfigParser
except:
    import ConfigParser

from .errors import *

logger = logging.getLogger(__name__)

# Get configuration directory
config_dir = user_config_dir(__name__.split(".")[0], False)

cid = "f7J0QNKkKcIkHMpk5lvq0wGZtBEpw1YcJ8Bea9Pv"
csid = "4OxQzkEnVc6OmTMKHgi77o8uFhEDz1QCuyKsNFm2Jo84r0TlGBXli0NrZ5uHTI769P3UurvDVB86awZkrEikCUJ7RM333BTS5oajO32BRwsNkPcOzA3JlBkPblvKM0l0"


class Warnings:
    insecure_warning = "You have chosen to disable certificate verification. Your connection is not secure!"


class Connection(object):
    """Connection class"""

    def __init__(
        self,
        server=None,
        username=None,
        password=None,
        oauth2_primary=True,
        api_key=None,
        debug=False,
        **kwargs,
    ):
        """Create Connection object

        Args:
            oauth2_primary (bool, required): oauth2 | api_key
            server (str, optional): Server URL or name in config file.
            username (str, optional)
            password (str, optional)
            api_key (str, optional)
            debug (bool, False): adds additional print statements to see url information
            **kwargs (optional): Mostly used when passing `server` as URL. See below.

        Keyword Args:
            insecure (bool): Whether or not to require valid SSL/TLS in requests
            path (str, optional): When specifying a server name instead of a URL, the path of the config file.
            server_name (str, optional): user-readable name of server. defaults to the server domain.
            url (str, optional): Alternative name for server argument.
            auth_url (str, optional): URL of the oauth server. Defaults to server url.

        """
        # Check url parameter to allow direct passing of ConnectionConfig instances

        server = server or kwargs.get("url")
        if re.match(r"^https?://", server):
            self.server_url = re.sub(r"([^/])$", r"\1/", server)
            self.server_name = re.sub(r"^https?://([A-Za-z0-9.\-_]+).*$", r"\1", server)
            self.auth_url = kwargs.pop("auth_url", self.server_url)
            self.client_id = cid
            self.client_secret = csid
            self.username = username
        else:
            path = kwargs.pop("path", "connect.cfg")
            config = ConnectionConfig(server)
            try:
                config.load_from_file(path)
            except ConfigParser.NoSectionError:
                config.create_config_object(path)

            self.server_name = config.server
            self.server_url = config.url
            self.auth_url = config.auth_url
            self.client_id = cid
            self.client_secret = csid
            self.username = username

        self.debug = debug
        self.token_url = self.auth_url.rstrip("/") + "/token/"
        self.max_retries = kwargs.pop("max_retries", 3)

        self._token_filename = os.path.join(
            config_dir, "{}_token.json".format(self.server_name)
        )

        try:
            self._token_last_updated = os.path.getmtime(self._token_filename)
        except OSError as e:
            if e.errno == errno.ENOENT:
                self._token_last_updated = None
            else:
                raise

        # Disable SSL check for localhost or when insecure=True
        self.insecure = kwargs.pop("insecure", False)
        if self.insecure:
            logger.error(Warnings.insecure_warning)
        if re.match(r"^https?://localhost", self.server_url) or self.insecure:
            requests.packages.urllib3.disable_warnings()
            os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
            self._verify = False
        else:
            os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = ""
            self._verify = True

        self._login(oauth2_primary, password, api_key)

    def _build_session(self, attempt=0):
        if attempt >= self.max_retries:
            raise LoginAttemptsExceededError()

        self.session = OAuth2Session(
            client=LegacyApplicationClient(client_id=self.client_id),
            auto_refresh_url=self.token_url,
            token_updater=self._token_updater,
        )
        self._set_session_retries()
        self.session.cookies.set_policy(BlockAll())

    def _clear_session(self):
        self._save_token(None)
        self.logout(self._token_filename)

    def _login(self, oauth2_primary, password=None, api_key=None, attempt=0):
        if password is not None:
            self._login_with_oauth2(password, attempt)
        elif api_key is not None:
            self._login_with_api_key(api_key, attempt)
        elif oauth2_primary:
            self._login_with_oauth2(password, attempt)
        else:
            self._login_with_api_key(api_key, attempt)

    def _login_with_api_key(self, api_key=None, isError=False, attempt=0):

        self._build_session(attempt=attempt)

        def request_api_key():
            if isError:
                print(
                    "Invalid API key. Input correct API Key, that you can generate in API Keys -> Generate API Key"
                )
            else:
                print(
                    "Connect to the cloud using API key, that you can generate in API Keys -> Generate API Key"
                )
            return input("Enter API token key:")

        # get token from json file
        if not api_key and os.path.isfile(self._token_filename):
            try:
                with open(self._token_filename, "r") as f:
                    apiKeyData = JSON.load(f)
            except Exception as e:
                self._clear_session()
            else:
                if apiKeyData:
                    api_key = apiKeyData.get("access_token")

        # request token from user
        if not api_key:
            api_key = request_api_key()

        # if API key is still empty exit
        if not api_key:
            return 0

        data = {
            "access_token": api_key,
            "token_type": "Bearer",
        }
        self.session.token = data

        try:
            self._save_token(self.session.token)
            response = self.request("get", "team-info/")
        except Exception as e:
            self._clear_session()
            self._login_with_api_key(isError=True, api_key=None, attempt=attempt + 1)
        else:
            if response.status_code != 200:
                self._clear_session()
                self._login_with_api_key(
                    isError=True, api_key=None, attempt=attempt + 1
                )
            else:
                print("Client connected to the cloud successfully")

    def _login_with_oauth2(self, password, attempt=0):

        self._build_session(attempt=attempt)

        if self.username is None or password is None:
            if os.path.exists(self._token_filename):
                with open(self._token_filename, "r") as f:
                    self.session.token = JSON.load(f)
                try:
                    self.session.refresh_token(
                        self.token_url,
                        auth=(self.client_id, self.client_secret),
                        verify=self._verify,
                    )
                    self._save_token(self.session.token)
                except:
                    os.remove(self._token_filename)
                    print("Invalid ClientID, Enter Username And Password to Login")
                    self.session.token = {}
                    self.username = input("Enter username: ")
                    password = getpass("Enter password: ")

            else:
                print(
                    "Connect to SensiML Cloud using your SensiML Account. Sign up for free at https://sensiml.com/plans/community-edition/"
                )
                self.username = input("Enter username: ")
                password = getpass("Enter password: ")

        if not self.session.token:
            try:
                r = self.session.fetch_token(
                    token_url=self.token_url,
                    username=self.username,
                    password=password,
                    # UAA endpoints always require id/secret as basic auth
                    auth=(self.client_id, self.client_secret),
                    verify=self._verify,
                )
            except (InvalidGrantError, MissingTokenError) as e:
                if isinstance(e, InvalidGrantError):
                    print("Invalid log in credentials, please try again.")
                else:
                    print(
                        "No token received: The remote server did not return a valid authorization response."
                    )
                if os.path.exists(self._token_filename):
                    os.remove(self._token_filename)
                password = None
                self.username = None
                return self._login_with_oauth2(password, attempt=attempt + 1)
            self._save_token(self.session.token)

        logger.debug("Login completed")

    def request(self, method, path, json=None, params=None, headers=None, **kwargs):
        """Sends request to server with authentication

        Args:
            method (str): HTTP Method.
            path (str): Path from server root to request.
            json (dict, optional): dictionary of information to be submitted as json
            headers (dict, optional): http headers to be sent
            **kwargs: Additional keyword arguments passed to underlying request library.

        Keyword Args:
            data (dict): Data to upload as application/x-www-form-urlencoded or multipart/form-data.
                Do not use in conjunction with json.
            paged (boolean): Return [PagedRequest] iterator instead of response

        Returns:
            Response: Response object from requests library with the result of the request
        """

        attempt = kwargs.pop("attempt", 0)
        paged = kwargs.pop("paged", False)
        verify = kwargs.pop("verify", self._verify)
        url = self.server_url.strip("/") + "/" + path.lstrip("/")
        if self._token_last_updated != os.path.getmtime(self._token_filename):
            with open(self._token_filename, "r") as f:
                self.session.token = JSON.load(f)

        while attempt < self.max_retries:

            if self.debug:
                print(
                    f"{ datetime.now(timezone.utc)} {method} {url} attempt: {attempt}"
                )

            if paged:
                return PagedRequest(
                    self.session,
                    method,
                    url,
                    json=json,
                    headers=headers,
                    params=params,
                    verify=verify,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    **kwargs,
                )
            else:
                response = self.session.request(
                    method,
                    url,
                    json=json,
                    params=params,
                    headers=headers,
                    verify=verify,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    **kwargs,
                )

            attempt += 1

            if self.debug:
                print(f"status code: {response.status_code}")

            if response.status_code in [504]:
                continue

            # temp workaround for bug in django-oauth-toolkit
            if response.status_code not in [401]:
                return response

            try:
                # If normal autorefresh fails we force it
                token = self.session.refresh_token(
                    self.token_url,
                    auth=(self.client_id, self.client_secret),
                    verify=self._verify,
                )
                self._save_token(token)
            except (TokenExpiredError, InvalidGrantError) as e:
                print("Refresh Token expired, please re-enter your login information.")
                try:
                    os.unlink(
                        os.path.join(
                            config_dir, "{}_token.json".format(self.server_name)
                        )
                    )
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
                self._login(None)

        return response

    def paged_request(
        self, method, path, json=None, headers=None, limit=1000, **kwargs
    ):
        """Returns an iterator that will request a new page each call.

        The result returned by the iterator will be the json decoded data of the
        'results' property of the response. Function args are the same as [request()]

        Args:
            method
            path
            json
            headers
            limit (Optional[int]): Number of entries per page, defaults to 1000.
                Note: this will override the 'limit' param specified in 'params'
                request kwarg.
            **kwargs: passed to request method
        """
        params = kwargs.pop("params", {})
        params["limit"] = limit
        return self.request(
            method, path, json, headers, params=params, paged=True, **kwargs
        )

    def file_request(
        self, url, file_path, data, read_type="rb", method="post", **kwargs
    ):
        """Loads a file and formats metadata for specific endpoints and makes server request

        Args:
            path (str): Path to API endpoint
            file_path (str): Local path to file
            data (dict): Data to send to server along with request. May be sent in nonstandard headers instead of request body for some endpoints.
            method (str, optional): HTTP method, defaults to POST
            read_type (str, optional): Mode to open the file. defaults to binary read ('rb')

        Returns:
            Response
        """
        with open(file_path, read_type) as f:

            if url.find("capture/") >= 0:
                request = partial(
                    self.request,
                    method,
                    url,
                    files={"file": f},
                    data=data,
                )
            elif url.find("featurefile/") >= 0:
                request = partial(
                    self.request,
                    method,
                    url,
                    files={"data_file": f},
                    data=data,
                )
            else:
                request = partial(
                    self.request, method, url, data=data, files={"file": f}
                )

            # call method with previously applied arguments
            return request(**kwargs)

    @staticmethod
    def logout(name):
        """Deletes saved session data for name"""
        try:
            os.unlink(os.path.join(config_dir, "{}_token.json".format(name)))
        except OSError as e:
            if e.errno == errno.ENOENT:
                print("No session saved for {}.".format(name))
                return
            else:
                raise
        print("Logged out.")

    def _set_session_retries(self):
        """set the number of retries a session has"""
        self.session.mount(
            "http://", requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        )
        self.session.mount(
            "https://", requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        )

    def _token_updater(self, token, save=True):
        """a simple function to save our token on refresh

        Args:
            token (dict): a token object for oauth2
        """
        self.session.token = token
        if save:
            self._save_token(token)

    def _save_token(self, token):
        if not os.path.exists(config_dir):
            try:
                os.makedirs(config_dir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
        with open(self._token_filename, "w") as f:
            # f.write(json.dumps(token))
            JSON.dump(token, f)
            self._token_last_updated = os.path.getmtime(f.name)


# With ouath we don't want to send cookies, this will overwite and block session cookies
class BlockAll(CookiePolicy):
    return_ok = (
        set_ok
    ) = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False


class PagedRequest(object):
    def __init__(self, session, method, path, json=None, headers=None, **kwargs):
        self.session = session
        self._method = method
        self._path = path
        self._json = json
        self._headers = headers
        self._kwargs = kwargs

        self.response = None

    def __iter__(self):
        return self

    def next(self):
        """
        Returns:
            (dict) content of the results property in the response
        """
        if not self.response:
            self.response = self.session.request(
                self._method, self._path, self._json, self._headers, **self._kwargs
            )
        else:
            json = self.response.json()
            path = json["next"]
            if not path:
                raise StopIteration
            self.response = self.session.request(
                self._method, path, self._json, self._headers, **self._kwargs
            )
        self.response.raise_for_status()
        return self.response.json()["results"]



from typing import Dict, List, Optional, Tuple, Type

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.lines import Line2D


def plot_segments(
    segments,
    capture=None,
    capture_name=None,
    labels=None,
    current_start=0,
    ylim=None,
    figsize: Tuple = (30, 8),
    currentAxis=None,
    xlim=None,
):

    if currentAxis is None:
        if capture is not None:
            currentAxis = capture.plot(figsize=figsize)
        else:
            plt.figure(figsize=figsize)
            currentAxis = plt.gca()

    cmap = plt.cm.rainbow

    if capture is not None:
        y_min = capture.describe().loc["min"].min()
        y_max = capture.describe().loc["max"].max()
        x_min = 0
        x_max = capture.shape[0]
    else:
        y_min = 0
        y_max = 1
        x_min = 0
        x_max = segments[-1].end

    if ylim:
        y_min = ylim[0]
        y_max = ylim[1]

    if labels is None:
        labels = segments.label_values

    delta = 1 / len(labels)
    label_float = np.arange(0, 1, delta)
    label_float[-1] = 1.0

    label_colors = {labels[index]: cmap(x) for index, x in enumerate(label_float)}
    label_legend = [Line2D([0], [0], color=cmap(x), lw=4) for x in label_float] + [
        Line2D([0], [0], color="white", lw=4)
    ]

    for index, seg in segments.iterrows():
        y_origin = y_min

        x_origin = seg["capture_sample_sequence_start"] + current_start
        x_final = (
            seg["capture_sample_sequence_end"] - seg["capture_sample_sequence_start"]
        )
        y_final = y_max - y_min

        currentAxis.add_artist(
            plt.Rectangle(
                (x_origin, y_origin),
                x_final,
                y_final,
                alpha=0.2,
                color=label_colors[seg["label_value"]],
            )
        )

    if xlim:
        currentAxis.set_xlim(left=xlim[0], right=xlim[1])
    else:
        currentAxis.set_xlim(left=x_min, right=x_max)

    currentAxis.legend(label_legend, labels + [""], loc=4, ncol=len(label_legend))
    """
    if capture_name:
        currentAxis.text(
            current_start,
            y_max,
            capture_name,
            style="italic",
        )
    """


def plot_threshold_space(
    capture,
    column,
    datasegments=None,
    labels=None,
    threshold_type="sum",
    threshold_width=25,
    xlim=None,
    figsize=(16, 4),
):
    M = []
    for i in range(capture.shape[0]):
        if threshold_type == "absolute sum":
            M.append(np.abs(capture[column].values[i : i + threshold_width]).sum())
        elif threshold_type == "sum":
            M.append(capture[column].values[i : i + threshold_width].sum())
        elif threshold_type == "std":
            M.append(capture[column].values[i : i + threshold_width].std())
        elif threshold_type == "variance":
            M.append(capture[column].values[i : i + threshold_width].var())
        else:
            print("Invalid threshold type")
            return

    test_capture = pd.DataFrame(M)
    if xlim is None:
        xlim = (0, len(M))
    ax = test_capture.plot(figsize=figsize, xlim=xlim)
    if datasegments is not None:
        datasegments.plot_segments(
            figsize=figsize,
            capture=test_capture,
            currentAxis=ax,
            labels=labels,
            xlim=xlim,
        )



import pandas as pd
from os.path import dirname
from os.path import join
import numpy as np


class ClassLabelException(Exception):
    pass


class DataSets:
    def __init__(self):
        pass

    def load_activity_raw(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, "data", "activity_raw.csv"))

    def load_activity_raw_toy(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, "data", "activity_raw_toy.csv"))

    def load_activity_norm(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, "data", "activity_norm.csv"))

    def load_activity_quant(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, "data", "activity_quant.csv"))

    def load_activity_quant_toy(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, "data", "activity_quant_toy.csv"))

    def load_gesture_raw(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, "data", "gesture_raw.csv"))

    def load_toy_data(self):
        module_path = dirname(__file__)
        return pd.read_csv(join(module_path, "data", "toy_data.csv"))

    @staticmethod
    def generate_step_data(
        window_size=100,
        num_classes=5,
        num_iterations=2,
        scale_factor=100,
        noise_scale=5,
        class_labels=None,
        class_values=None,
        loop=False,
    ):
        """Generate a labeled test data vector.

        Args:
            window_size (int, optional): The size of each segment window.
            num_classes (int, optional): The number of separate classes
            num_iterations (int, optional): The number of iterations to create.
            scale_factor (int, optional): scale the values of the data by this much before adding noise
            noise_scale (int, optional): The range of noise to add (-noise_scale:noise_scale)
            class_labels (None, optional): you can pass specific class labels as an array, otherwise defaults to
             the range(num_classes)
            loop (bool, False): if True starts from class 1 at end of iteration. if False, starts at class num_classes and reverses at end of iteration

        Returns:
            DataFrame: a test data vector for building and testing models
        """
        class_labels = check_class_label(num_classes, class_labels)
        class_values = check_class_values(num_classes, class_values)
        delta = window_size * num_classes
        signal_length = num_iterations * delta

        y = np.zeros(signal_length)
        y_label = np.zeros(signal_length)

        for j in range(num_iterations):
            y_temp = np.zeros(window_size * num_classes)
            if j % 2 == 0:
                for i in range(num_classes):

                    y_temp[window_size * i : window_size * (i + 1)] = class_values[i]
                    y_label[
                        window_size * i + j * delta : window_size * (i + 1) + j * delta
                    ] = class_labels[i]
            else:
                for i in range(num_classes):
                    index = -1 - i
                    class_index = num_classes + index
                    if loop:
                        index = i
                        class_index = index
                    y_temp[window_size * i : window_size * (i + 1)] = class_values[
                        index
                    ]
                    y_label[
                        window_size * i + j * delta : window_size * (i + 1) + j * delta
                    ] = class_labels[class_index]
            y[delta * j : delta * (j + 1)] = y_temp * scale_factor

        if noise_scale > 0:
            y += np.array(
                [
                    np.random.choice(range(-noise_scale, noise_scale))
                    for i in range(len(y))
                ]
            )

        return pd.DataFrame({"Data": y.astype(np.int16), "Label": y_label.astype(int)})

    @staticmethod
    def generate_harmonic_data(
        window_size=100,
        num_classes=5,
        num_iterations=2,
        scale_factor=100,
        noise_scale=5,
        class_labels=None,
        harmonic_nodes=None,
    ):
        """Generate a labeled test data sin vector.

        Args:
            window_size (int, optional): The size of each segment window.
            num_classes (int, optional): The number of separate classes
            num_iterations (int, optional): The number of iterations to create.
            scale_factor (int, optional): scale the values of the data by this much before adding noise
            noise_scale (int, optional): The range of noise to add (-noise_scale:noise_scale)
            class_labels (None, optional): you can pass specific class labels as an array, otherwise defaults to
             the range(num_classes)

        Returns:
            DataFrame: a test data vector for building and testing models
        """

        class_labels = check_class_label(num_classes, class_labels)
        harmonic_nodes = check_class_values(num_classes, harmonic_nodes)

        delta = window_size * num_classes
        signal_length = num_iterations * delta

        y = np.zeros(signal_length)
        x = np.arange(signal_length)

        y_label = np.zeros(signal_length)
        for j in range(num_iterations):
            y_temp = np.zeros(delta)
            x = np.arange(window_size)
            if j % 2 == 0:
                for i in range(num_classes):
                    y_temp[window_size * i : window_size * (i + 1)] = get_standing_wave(
                        x, window_size, harmonic_nodes[i]
                    )
                    y_label[
                        window_size * i + j * delta : window_size * (i + 1) + j * delta
                    ] = class_labels[i]
            else:
                for i in range(num_classes):
                    y_temp[window_size * i : window_size * (i + 1)] = get_standing_wave(
                        x, window_size, harmonic_nodes[-1 - i]
                    )
                    y_label[
                        window_size * i + j * delta : window_size * (i + 1) + j * delta
                    ] = class_labels[num_classes - 1 - i]
            y[delta * j : delta * (j + 1)] = y_temp * scale_factor

        if noise_scale > 0:
            y += np.array(
                [
                    np.random.choice(range(-noise_scale, noise_scale))
                    for i in range(len(y))
                ]
            )

        return pd.DataFrame({"Data": y.astype(np.int16), "Label": y_label.astype(int)})


def get_standing_wave(x, L, n):
    ld = 2.0 * L / n
    return np.sin(x * (2 * np.pi) / ld)


def check_class_label(num_classes, class_labels=None):
    if class_labels is None:
        class_labels = range(1, num_classes + 1)
    elif len(class_labels) != num_classes:
        raise ClassLabelException(
            "The number of classes is not equal to the number of class labels."
        )

    for i in class_labels:
        if type(i) != type(1):
            raise ClassLabelException("class labels must be integers.")
    if i <= 0:
        raise ClassLabelException("class labels must be greater than zero.")

    return class_labels


def check_class_values(num_classes, class_values=None):
    if class_values is None:
        class_values = range(1, num_classes + 1)
    elif len(class_values) != num_classes:
        raise ClassLabelException(
            "The number of classes is not equal to the number of class values."
        )

    for i in class_values:
        if type(i) != type(1):
            raise ClassLabelException("class values must be type int")
        if i <= 0:
            raise ClassLabelException("class values must be greater than zero.")

    return class_values



import os
import json
from sensiml.base import utility


class Team(object):
    """Base class for a transform object"""

    _uuid = ""
    _name = None

    def __init__(self, connection):
        self._connection = connection

    def get_user(self):
        """Get Information about the users on your team."""

        url = "user/"
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        return response_data

    def team_subscription(self):
        """Get Information about your teams subscription."""
        url = "team-subscription/"
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        return response_data

    def team_info(self):
        """Get information about your specific team."""
        url = "team-info/"
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        return response_data



import json
import logging
import os
import re
from ast import Str
from concurrent import futures
from concurrent.futures import ThreadPoolExecutor
from tempfile import NamedTemporaryFile
from typing import Dict, List, Optional, Tuple
from xml.dom import ValidationErr

import requests
from appdirs import user_config_dir
from packaging import version

try:
    from qgrid import show_grid
except:
    pass

from time import sleep

import pandas as pd
from pandas import DataFrame

import sensiml.base.utility as utility
from sensiml.base.snippets import Snippets, function_help
from sensiml.connection import Connection
from sensiml.datamanager import (
    ClientPlatformDescriptions,
    Functions,
    Projects,
    SegmenterSet,
    Team,
)
from sensiml.datamanager.captures import CaptureExistsError
from sensiml.datamanager.knowledgepack import delete_knowledgepack, get_knowledgepack
from sensiml.datasets import DataSets
from sensiml.dclproj.upload import upload_project, upload_project_dcli

# from sensiml.base.exceptions import *
from sensiml.pipeline import Pipeline

config_dir = user_config_dir(__name__.split(".")[0], False)

SERVER_URL = "https://sensiml.cloud/"

AUTH2_PRIMARY = True

logger = logging.getLogger("SensiML")

__version__ = "2023.2.0"


def project_set(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        if self._project is None:
            print("Project must be set.")
            return None
        else:
            return func(*args, **kwargs)

    return wrapper


def print_list(func):
    """This is a wrapper for printing out lists of objects stored in SensiML Cloud"""

    def wrapper(*args, **kwargs):

        result = func(*args, **kwargs)

        if len(result.keys()) == 0 and kwargs.get("silent", False) == False:
            print(
                "No {} stored on SensiML Cloud for this project.".format(
                    " ".join(func.__name__.split("_")[1:])
                )
            )
            return None
        if kwargs.get("get_objects", False) is True:
            return result

        tmp_dict = [
            {
                "Name": name,
                "UUID": item.uuid,
                "Created": item.created_at,
                "Last Modified": item.created_at
                if not hasattr(item, "last_modified")
                else item.last_modified,
            }
            for name, item in result.items()
        ]

        extra = {}
        additional_keys = []
        if func.__name__ == "list_queries":
            extra = [
                {"Cached": True if item.cache else ""} for name, item in result.items()
            ]
            additional_keys = ["Cached"]

        if func.__name__ == "list_sandboxes":
            extra = [
                {"CPU Used (seconds)": item.cpu_clock_time}
                for name, item in result.items()
            ]
            additional_keys = ["CPU Used (seconds)"]

        for index, value in enumerate(extra):
            tmp_dict[index].update(value)

        return DataFrame(
            tmp_dict,
            columns=["Name", "Last Modified", "Created", "UUID"] + additional_keys,
        )

    return wrapper


class Client(object):
    def __init__(
        self,
        server=SERVER_URL,
        path="connect.cfg",
        use_jedi=False,
        insecure=False,
        skip_validate=False,
        verbose_connection=False,
        **kwargs,
    ):

        self._project = None
        self._pipeline = None
        auth_url = server + "oauth/"

        self._connection = Connection(
            server=server,
            auth_url=auth_url,
            path=path,
            insecure=insecure,
            oauth2_primary=AUTH2_PRIMARY,
            debug=verbose_connection,
            **kwargs,
        )

        if not self.validate_client_version(skip_validate):
            return
        self.projects = Projects(self._connection)
        self.datasets = DataSets()
        self.functions = Functions(self._connection)
        self.platforms_v2 = ClientPlatformDescriptions(self._connection)
        self.platforms = self.platforms_v2
        self.team = Team(self._connection)
        self.snippets = Snippets(
            self.list_functions(kp_functions=False, qgrid=False),
            self.functions.function_list,
        )
        self._feature_files = None

        if use_jedi is False:
            self.setup_jedi_false()

    def setup_jedi_false(self):
        """This is a temporary bug fix in ipython autocomplete"""
        try:
            mgc = get_ipython().magic
            mgc("%config Completer.use_jedi = False")
        except:
            pass

    def validate_client_version(self, skip_validate=True):
        """Perform a Validation check to see if this version of SensiML is up to date with the latest."""

        if skip_validate:
            return True

        url = "version/"

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        if response_data.get("SensiML_Python_Library_Windows_Minimum") is None:
            return True

        if version.parse(
            response_data["SensiML_Python_Library_Windows_Minimum"]
        ) > version.parse(__version__):
            print(
                f"Error: The SensiML Python SDK is out of date.\n\n\tCurrent installed version = {__version__}. \n\tMinimum supported version = {response_data['SensiML_Python_Library_Windows_Minimum']}."
            )
            print(
                "\nTo update the SensiML Python SDK run \n\n\tpip install sensiml -U\n\nIf you are in a notebook, you can execute the following in a cell then restart the shell.\n\n\t!pip install sensiml -U"
            )
            print(
                "\nTo disable this validation check and connect anyway you can use the skip_validate parameter\n\n\tclient = Client(skip_validate=True)"
            )
            return False

        return True

    def logout(self, name=None):
        """Logs out of the current connection."""
        if name is None:
            name = self._connection.server_name

        Connection.logout(name)

    def get_url(self, url):

        response = self._connection.request("get", url)
        print(response.json())
        return response

    def account_info(self):
        """Get information about your account Usage"""

        url = "team-info/"
        return self.get_url(url)

    def account_subscription(self):
        """Get information about your account Subscription"""

        url = "team-subscription/"
        return self.get_url(url)

    def get_function(self, name):
        """Gets a function method call"""
        return self.functions.function_list[name]

    def function_description(self, name):
        """Gets a description of a pipeline function."""
        print(self.functions.create_function_call(name).__doc__)

    def function_help(self, name):
        """Prints a shortened description of a function."""
        print(function_help(self.functions.function_list[name]))

    def list_functions(
        self, functype=None, subtype=None, kp_functions=False, qgrid=False
    ):
        """Lists all of the functions available on SensiML Cloud

        Returns:
            Dataframe

        Args:
            functype (str, None): Return only functions with the specified type. ie. "Segmenter"
            subtype (str, None): Return only functions with the specified subtype. ie. "Sensor"
            kp_functions (boolean, True): Return only functions that run on the loaded device.
            Excludes functions such as feature selection and model training.
        """

        df = (
            DataFrame(
                [
                    {
                        "NAME": f.name,
                        "TYPE": f.type,
                        "DESCRIPTION": f.description.lstrip("\n").lstrip(" ")
                        if f.description
                        else "",
                        "SUBTYPE": f.subtype,
                        "KP FUNCTION": f.has_c_version,
                        "UUID": f.uuid,
                        "LIBRARY": f.library_pack,
                        "AVAILABLE": f.automl_available,
                    }
                    for name, f in self.functions.function_list.items()
                ]
            )
            .sort_values(by=["TYPE", "SUBTYPE"])
            .reset_index(drop=True)[
                [
                    "NAME",
                    "TYPE",
                    "SUBTYPE",
                    "DESCRIPTION",
                    "KP FUNCTION",
                    "AVAILABLE",
                    "UUID",
                    "LIBRARY",
                ]
            ]
        )

        if functype:
            df = df[df["TYPE"] == functype]

        if subtype:
            df = df[df["SUBTYPE"] == subtype]

        if kp_functions:
            df = df[df["KP FUNCTION"] == True][
                ["NAME", "TYPE", "SUBTYPE", "DESCRIPTION", "LIBRARY"]
            ]

        if qgrid:
            return show_grid(df.reset_index(drop=True))
        else:
            return df.reset_index(drop=True)

    def delete_project(self):
        """Deletes a project"""
        if self._project is not None:
            self._project.delete()

    @print_list
    def list_projects(self, get_objects=False, silent=False):
        """Lists all of the projects on SensiML Cloud

        Returns:
            DataFrame: projects on SensiML Cloud
        """
        return self.projects.build_project_dict()

    def list_segmenters(
        self,
    ):
        if self._project is None:
            print("project must be set to list segmenters.")
            return None

        segmenters = SegmenterSet(self._connection, self._project)

        if not len(segmenters.objs):
            print("No segmenters stored on the Cloud.")
            return None

        return segmenters.to_dataframe()

    def get_knowledgepack(self, uuid, get_flatbuffer=True):
        """Retrieve a Knowledge Pack by uuid from the server associated with current project

        Args:
            uuid (str): unique identifier for Knowledge Pack

        Returns:
            TYPE: a Knowledge Pack object
        """

        return get_knowledgepack(uuid, self._connection, get_flatbuffer=get_flatbuffer)

    def delete_knowledgepack(self, uuid):
        """Delete Knowledge Pack by uuid from the server associated with current project

        Args:
            uuid (str): unique identifier for Knowledge Pack

        Returns:
            TYPE: a Knowledge Pack object
        """

        return delete_knowledgepack(uuid, self._connection)

    @property
    def project(self):
        """The active project"""
        return self._project

    def get_featurefile(self, uuid):
        """Get a FeatureFile by uuid

        Args:
            get_objects (bool, False): Also return the FeatureFile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._feature_files.get_featurefile(uuid)

    def get_datafile(self, uuid):
        """Get a datafile by uuid

        Args:
            get_objects (bool, False): Also return the datafile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._feature_files.get_featurefile(uuid)

    @print_list
    def list_featurefiles(self, get_objects=False, silent=False):
        """List all feature and data files for the active project.

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._feature_files.build_featurefile_list()

    @print_list
    def list_datafiles(self, get_objects=False, silent=False):
        """List all feature and data files for the active project.

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._feature_files.build_datafile_list()

    @print_list
    def list_captures(self, get_objects=False, force=False):
        """List all captures for the active project

        Args:
            get_objects (bool, False): Also return the capture objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._captures.build_capture_list(force=force)

    @print_list
    def refresh_capture_list(self, get_objects=False):
        """Refresh the internal capture list from the server

        Args:
            get_objects (bool, False): Also return the capture objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._captures.build_capture_list(force=True)

    @project_set
    def get_captures(self, force: bool = True):
        """Returns the capture set object"""
        return self._project._captures.build_capture_list(force=force)

    def download_capture(
        self, filename: str, outdir: Optional[str] = None, force: bool = False
    ):
        """Downloads the capture file from the server

        Args:
            filename (str): name of the file to download
            outdir (Optional[str], optional): directory to put the downloaded file, defaults to current directory.
            force (bool, optional): refresh the capture list

        """
        return self._project._captures.build_capture_list(force=force)[
            filename
        ].download(outdir=outdir)

    def _download_captures(
        self,
        filenames: List[str],
        outdir: Optional[str] = None,
        expires_in: int = 100,
        verbose=True,
    ):

        capture_response = self.project.captures.get_capture_urls_by_name(
            filenames, expires_in=expires_in
        )

        for capture in capture_response:
            url = capture["url"]

            if capture.get("local"):
                response = self._connection.request("get", url)
            else:
                response = requests.get(url)

            if outdir is None:
                outdir = "./"

            if capture["name"][-4:] == ".csv":
                with open(os.path.join(outdir, capture["name"]), "w") as out:
                    out.write(response.text)
                    if verbose:
                        print(
                            f"Capture saved to {os.path.join(outdir, capture['name'])}"
                        )

            elif capture["name"][-4:] == ".wav":
                with open(os.path.join(outdir, capture["name"]), "wb") as out:
                    out.write(response.content)
                    if verbose:
                        print(
                            f"Capture saved to {os.path.join(outdir, capture['name'])}"
                        )

            else:
                raise Exception("Unknown file type")

        return "Success"

    def _download_capture_parallel_multithreading_urls(
        self,
        filenames: List[str],
        outdir: Optional[str] = None,
        batch_size: int = 10,
        max_workers: int = 12,
        expires_in=300,
    ):
        batch_size = batch_size
        file_batches = []
        last = 0
        for i in range(0, len(filenames), batch_size):
            file_batches.append(filenames[i : i + batch_size])
            last = i + batch_size
        file_batches.append(filenames[last:])

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            future_to_key = {
                executor.submit(
                    self._download_captures, batch, outdir, expires_in, False
                ): batch
                for batch in file_batches
            }

            for future in futures.as_completed(future_to_key):
                key = future_to_key[future]
                exception = future.exception()

                if not exception:
                    yield key, future.result()
                else:
                    yield key, exception

    def download_captures(
        self,
        filenames: List[str],
        outdir: Optional[str] = None,
        max_workers: int = 8,
        batch_size=10,
    ):
        """Downloads a list of capture files from the server

        Args:
            filenames (List[str]): List of file names to download
            outdir (Optional[str], optional): Directory to download files to. Defaults to None.
            max_workers (int, optional): number of worker threads to use. Defaults to 8.
            batch_size (int, optional): batch size to use in each worker thread. Defaults to 10.
        """
        count = 0
        for key, result in self._download_capture_parallel_multithreading_urls(
            filenames, outdir=outdir, max_workers=max_workers, batch_size=batch_size
        ):
            count += batch_size
            print(f"{count}/{len(filenames)} {key} result: {result}")

    def download_project(
        self,
        outdir: Optional[str] = None,
        max_workers: int = 8,
        batch_size=10,
    ):
        """Downloads all of the capture files along with a .dcli file which includes segment and metadata information for each file.

        Args:
            outdir (Optional[str], optional): Directory to store project. Defaults to project name.
            max_workers (int, optional): Number of parallel workers to use for downloading files. Defaults to 8.
            batch_size (int, optional): The batch size to use for each worker. Defaults to 10.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        if outdir is None:
            outdir = self.project.name

        if not os.path.isdir(outdir):
            os.mkdir(outdir)

        print(f"Downloading Project to {outdir}")
        dcli = self.project.get_project_dcli()
        json.dump(dcli, open(os.path.join(outdir, f"{self.project.name}.dcli"), "w"))
        captures = self.list_captures().Name.values
        self.download_captures(
            captures, outdir=outdir, max_workers=max_workers, batch_size=batch_size
        )
        print(f"Download Complete")

    @print_list
    def list_capture_configurations(self, get_objects=False):
        """List all captures for the active project

        Args:
            get_objects (bool, False): Also return the capture objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._capture_configurations.build_capture_list()

    def capture_configurations(self):
        """Returns the capture set object"""
        return self._project._capture_configurations.build_capture_list()

    @print_list
    def list_sandboxes(self, get_objects=False):
        """List all sandboxes for the active project.

        Args:
            get_objects (bool, False): Also return the sandbox objects.

        """
        print("list_sanboxes has been deprecated in favor of list_pipelines.")
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._sandboxes.build_sandbox_list()

    @project_set
    def get_pipelines(self):
        """Returns the pipeline set dictionary"""
        return self._project._sandboxes.build_sandbox_list()

    @print_list
    def list_pipelines(self, get_objects=False):
        """List all pipelines for the active project.

        Args:
            get_objects (bool, False): Also return the sandbox objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._sandboxes.build_sandbox_list()

    @print_list
    def list_queries(self, get_objects=False):
        """List all queries for the active project.

        Args:
            get_objects (bool, False): Also return the query objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._queries.build_query_list()

    @project_set
    def get_queries(self):
        """Returns the query set dictionary"""
        return self._project._queries.build_query_list()

    @project.setter
    def project(self, name):
        self._project = self.projects.get_or_create_project(name)

    @property
    def pipeline(self):
        """The active pipeline"""
        return self._pipeline

    @pipeline.setter
    def pipeline(self, name):
        if self._project is None:
            raise Exception("Project must be set before a pipeline can be created")

        self._pipeline = Pipeline(self, name=name)

    def create_query(
        self,
        name: str,
        columns: list = [],
        metadata_columns: list = [],
        metadata_filter: str = "",
        segmenter=None,
        label_column: str = "",
        combine_labels=None,
        force: bool = False,
        renderer=None,
        capture_configurations: str = "",
    ):
        """Create a query to use as input data in a pipeline.

        Args:
            name (str): Name of the query.
            columns (list, optional): Columns to add to the query result.
            metadata_columns (list, optional): Metadata to add to the query result.
            metadata_filter (str, optional): Filter to apply to the query.
            segmenter (int, optional): Segmenter to filter query by.
            force (bool, False): If True overwrite the query on kb cloud.

        Returns:
            object: Returns a query object that was created.
        """

        query = self.project.queries.get_query_by_name(name)

        new = False
        if query is not None and not force:
            raise Exception("Query already exists. Set force=True to overwrite.")
        elif query is not None and force:
            query.columns.clear()
            query.metadata_columns.clear()
            query.metadata_filter = ""
            query.label_column = ""
            query.capture_configurations = ""
            query.segmenter = None

        else:
            query = self.project.queries.new_query()
            query.name = name
            new = True

        for col in columns:
            logger.debug("query_column:" + str(col))
            query.columns.add(col)

        for col in metadata_columns:
            logger.debug("query_metadata_column:" + str(col))
            query.metadata_columns.add(col)

        if metadata_filter:
            logger.debug("query_metadata_filter:" + str(metadata_filter))
            query.metadata_filter = metadata_filter

        if label_column:
            query.label_column = label_column

        if combine_labels:
            query.combine_labels = combine_labels

        if capture_configurations:
            query.capture_configurations = capture_configurations

        query.metadata_columns.add("segment_uuid")

        if isinstance(segmenter, str):
            segmenters = self.list_segmenters()
            segmenter_id = segmenters[segmenters["name"] == segmenter].id
            if segmenter_id.shape[0] != 1:
                raise Exception("Segmenter {} not found".format(segmenter))

            query.segmenter = int(segmenter_id.values[0])
        else:
            query.segmenter = segmenter

        if new:
            query.insert(renderer=renderer)
        else:
            query.update(renderer=renderer)

        return query

    def get_query(self, name):
        if self.project is None:
            print("Project must be set first")
            return

        return self.project.queries.get_query_by_name(name)

    def upload_data_file(
        self, name: str, path: str, force: bool = False, is_features: bool = False
    ):
        """Upload a .CSV file as either a FeatureFile or DataFile to the server.

        FeatureFiles are a collection of feature vectors and can be used in any step after the feature generation step
        DataFiles include sensor data and metadata and are used in any step prior to feature generation

        Args:
            name (str): Name of the file when it is uploaded
            path (str): The path to the file to upload
            force (bool, optional): Will overwrite if already exists. Defaults to False.
            is_features (bool, optional): If True, will upload as a feature file, if False will upload as a DataFile. Defaults to False.

        Returns:
            response: The response as a request object
        """
        logger.debug("set_feature_file:" + name + ":" + path)
        print('Uploading file "{}" to SensiML Cloud.'.format(name))
        if name[-4:] != ".csv":
            name = "{}.csv".format(name)

        feature_file = self._project._feature_files.get_by_name(name)
        if feature_file is None:
            new = True
            feature_file = self._project.featurefiles.new_featurefile()
        else:
            new = False
            if not force:
                raise Exception(
                    "A file with this name already exists. Use force=True to override"
                )

        feature_file.filename = name
        feature_file.is_features = is_features
        feature_file.path = path
        if new:
            return feature_file.insert()
        else:
            return feature_file.update()

    def upload_dataframe(
        self,
        name: str,
        dataframe: DataFrame,
        force: bool = False,
        is_features: bool = False,
    ):
        """Upload a pandas DataFrame as either a FeatureFile or DataFile to the server.

        FeatureFiles are a collection of feature vectors and can be used in any step after the feature generation step
        DataFiles include sensor data and metadata and are used in any step prior to feature generation

        Args:
            name (str): Name of the file when it is uploaded
            dataframe (DatFrame): Pandas DataFrame
            force (bool, optional): Will overwrite if already exists. Defaults to False.
            is_features (bool, optional): If True, will upload as a feature file, if False will upload as a DataFile. Defaults to False.

        Returns:
            response: The response as a request object
        """

        logger.debug("set_data:" + name)

        with NamedTemporaryFile(delete=False) as temp:
            dataframe.to_csv(temp.name, index=False)
            logger.debug("set_dataframe:" + name + ":" + temp.name)
            result = self.upload_data_file(
                name, temp.name, force=force, is_features=is_features
            )

        os.remove(temp.name)

        return result

    def upload_sensor_dataframe(
        self, name: str, dataframe: DataFrame, force: bool = False
    ):
        """Upload a pandas DataFrame as a DataFile to the server.

        DataFiles include sensor data and metadata and can be used in any any step prior to feature generation

        Args:
            name (str): Name of the file when it is uploaded
            dataframe (DatFrame): Pandas DataFrame
            force (bool, optional): Will overwrite if already exists. Defaults to False.

        Returns:
            response: The response as a request object
        """
        logger.debug("set_data:" + name)

        with NamedTemporaryFile(delete=False) as temp:
            dataframe.to_csv(temp.name, index=False)
            logger.debug("set_dataframe:" + name + ":" + temp.name)
            result = self.upload_data_file(
                name, temp.name, force=force, is_features=False
            )

        os.remove(temp.name)

        return result

    def upload_feature_dataframe(
        self, name: str, dataframe: DataFrame, force: bool = False
    ):
        """Upload a pandas DataFrame as a FeatureFile to the server.

        FeatureFiles are a collection of feature vectors and can be used in any step after the feature generation step

        Args:
            name (str): Name of the file when it is uploaded
            dataframe (DatFrame): Pandas DataFrame
            force (bool, optional): Will overwrite if already exists. Defaults to False.

        Returns:
            response: The response as a request object
        """
        logger.debug("set_data:" + name)

        with NamedTemporaryFile(delete=False) as temp:
            dataframe.to_csv(temp.name, index=False)
            logger.debug("set_dataframe:" + name + ":" + temp.name)
            result = self.upload_data_file(
                name, temp.name, force=force, is_features=True
            )

        os.remove(temp.name)

        return result

    def clear_session_cache(self):
        for _, _, filenames in os.walk(config_dir):
            for filename in filenames:
                if re.match(r"_token.json$", filename):
                    os.unlink(filename)

    def get_feature_statistics_results(self, query_name: str):
        query = self._project.queries.get_query_by_name(query_name)

        for i in range(100):
            results = query.get_feature_statistics()

            if results.get("results", None):
                return pd.DataFrame(results["results"])
            sleep(5)

        print("Not able to reach the results in the expected time. ")
        return results

    def upload_project(self, name: str, dclproj_path: str):
        """Upload a .dclproj file to the server

        Args:
            name (str): name of the project to create
            dclproj_path (str): path to the .dclproj file
        """
        if name in self.list_projects()["Name"].values:
            print("Project with this name already exists.")
            return

        upload_project(self, name, dclproj_path)

    def upload_project_dcli(self, name: str, dcli_path: str, force: bool = False):
        """Upload a .dcli project to the server

        Args:
            name (str): name of the project to create
            dcli_path (str): path to the .dclproj file
            force (bool): attempt to upload to an existing project
        """
        if name in self.list_projects()["Name"].values and not force:
            print(
                "Project with this name already exists. Set force=True to upload dcli data to this project."
            )
            return

        upload_project_dcli(self, name, dcli_path)

    def list_background_captures(self, get_objects=False):
        """List all captures for the active project

        Args:
            get_objects (bool, False): returning the list of capture objects.

        Return:
            pandas.DataFrame: if get_objects is False
            list of capture objects if get_objects is True
        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        back_captures = self._project._background_captures.get_background_captures()
        if get_objects:
            return back_captures

        tmp_dict = [
            {
                "filename": item._filename,
                "sample_rate": item._set_sample_rate,
                "number_samples": item._number_samples,
                "uuid": item._uuid,
                "last_modified": item._last_modified,
            }
            for item in back_captures
        ]

        return DataFrame(
            tmp_dict,
            columns=[
                "filename",
                "sample_rate",
                "number_samples",
                "uuid",
                "last_modified",
            ],
        )

    def download_background_capture_by_uuid(
        self, uuid: str, output_filename: str = "", output_folder="./"
    ):
        """Download a background capture providing its UUID.

        Args:
            uuid (str): UUID of the background capture
            output_filename (str: ""): Output filename. The original name of the the capture file is used if not provided.
            output_folder (str: "./"): The local folder path to store the downloaded file

        Return:
            server response
        """

        back_capture = self._project._background_capture
        back_captures = self._project._background_captures
        back_capture._uuid = uuid

        if output_filename == "":
            back_captures_info_list = back_captures.get_background_capture_urls_by_uuid(
                [uuid]
            )
            if len(back_captures_info_list) == 1:
                output_filename = back_captures_info_list[0]["name"]
            else:
                raise Exception("Capture file not found !")

        back_capture._filename = output_filename

        try:
            response = back_capture.download(outdir=output_folder)
        except Exception as e:
            print(e)

        return response



import json
from sensiml.datamanager.capture_configuration import CaptureConfiguration
import sensiml.base.utility as utility


class CaptureConfigurationExistsError(Exception):
    """Base class for a Capture exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class CaptureConfigurations:
    """Base class for a collection of Captures."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def create_capture_configuration(self, name, configuration, uuid=None):
        """Creates a capture object from the given filename and filepath.

        Args:
            name (str): desired name of the file on the server
            configuration (dict): configuration file
            uuid (str): Specify the uuid this configuration will be created with

        Returns:
            capture_configuration object

        Raises:
            CaptureClonfigurationExistsError, if the Capture already exists on the server
        """
        data = {"connection": self._connection, "project": self._project, "name": name}

        if self.get_capture_configuration_by_name(name) is not None:
            raise CaptureConfigurationExistsError(
                "capture configuration {0} already exists.".format(name)
            )
        else:
            instance = CaptureConfiguration.initialize_from_dict(data)
            instance.configuration = configuration
            instance.uuid = uuid
            instance.insert()

        return instance

    def build_capture_list(self):
        """Populates the function_list property from the server."""
        capture_list = {}

        response = self.get_capture_configurations()
        for capture_configuration in response:
            capture_list[capture_configuration.name] = capture_configuration

        return capture_list

    def get_capture_configuration_by_name(self, name):
        """Gets a capture configuration object from the server using its name property.

        Args:
            name (str): the capture configuration's name

        Returns:
            capture configuration  object or None if it does not exist
        """
        capture_list = self.get_capture_configurations()
        for capture in capture_list:
            if capture.name == name:
                return capture
        return None

    def __getitem__(self, key):
        if type(key) == str:
            return self.get_capture_configuration_by_filename(key)
        else:
            return self.get_capture_configurations()[key]

    def _new_capture_configuration_from_dict(self, data):
        """Creates a new capture using the dictionary.

        Args:
            data (dict): dictionary of capture properties

        Returns:
            capture configuration object
        """
        data.update({"connection": self._connection, "project": self._project})
        return CaptureConfiguration.initialize_from_dict(data)

    def get_capture_configurations(self):
        """Gets all capture configurations from the server.

        Returns:
            list of capture configurations for the project
        """
        # Query the server and get the json
        url = "project/{0}/captureconfiguration/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        # Populate each capture from the server
        data = []
        if err is False:
            for params in response_data:
                data.append(self._new_capture_configuration_from_dict(params))

        return data



from sensiml.method_calls.selectorcall import SelectorCall
from sensiml.datamanager.pipeline import PipelineStep

# AugmentationCallSet
class AugmentationCallSet(PipelineStep):
    """The base class for a collection of augmentation calls"""

    def __init__(self, name=""):
        super(AugmentationCallSet, self).__init__(
            name=name, step_type="AugmentationCallSet"
        )
        self._augmentations = []
        self._input_data = ""
        self._label_column = None
        self._passthrough_columns = None
        self._group_columns = None
        self._percent = 0.1

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def input_data(self):
        return self._input_data

    @input_data.setter
    def input_data(self, value):
        self._input_data = value

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._label_column = value

    def add_augmentation_call(self, *augmentations):
        """Adds one or more augmentation calls to the collection.

        Args:
            augmentation (AugmentationCall or list[AugmentationCall]): object(s) to append
        """
        for augmentation in augmentations:
            self._augmentations.append(augmentation)

    def remove_augmentation_call(self, *augmentations):
        """Removes one or more augmentation call from the collection.

        Args:
            augmentations (AugmentationCall or list[AugmentationCall]): object(s) to remove
        """
        for augmentation in augmentations:
            self._augmentations = [f for f in self._augmentations if f != augmentation]

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def passthrough_columns(self):
        return self._passthrough_columns

    @passthrough_columns.setter
    def passthrough_columns(self, value):
        self._passthrough_columns = value

    @property
    def group_columns(self):
        return self._passthrough_columns

    @group_columns.setter
    def group_columns(self, value):
        self._group_columns = value

    @property
    def percent(self):
        return self._percent

    @percent.setter
    def percent(self, value):
        self._percent = value

    def _to_list(self):
        gencalls = []
        for item in self._augmentations:
            gencalls.append(item._to_dict())
        return gencalls

    def _to_dict(self):
        selcalls_set = []
        set_dict = {}

        for item in self._augmentations:
            selcalls_set.append(item._to_dict())

        set_dict["type"] = "augmentationset"
        set_dict["name"] = getattr(self, "_name")
        set_dict["set"] = selcalls_set
        set_dict["inputs"] = {}

        set_dict["inputs"]["passthrough_columns"] = getattr(
            self, "_passthrough_columns"
        )
        set_dict["inputs"]["label_column"] = getattr(self, "_label_column")

        set_dict["inputs"]["group_columns"] = getattr(self, "_group_columns")

        set_dict["inputs"]["input_data"] = getattr(self, "_input_data")
        set_dict["label_column"] = getattr(self, "_label_column")

        set_dict["outputs"] = self._outputs

        return set_dict



import json
import os

from sensiml.base import utility
from sensiml.datamanager.base import Base, BaseSet


class FoundationModel(Base):
    """Base class for a transform object"""

    _uuid = ""
    _name = None

    _fields = [
        "uuid",
        "name",
        "features_count",
        "model_size",
        "created_at",
        "knowledgepack_description",
        "last_modified",
    ]

    _read_only_fields = [
        "uuid",
        "name",
        "features_count",
        "model_size",
        "created_at",
        "knowledgepack_description",
        "last_modified",
    ]

    _field_map = {"transform_type": "type"}

    def __init__(self, connection, uuid=None):
        self._connection = connection
        if uuid:
            self.uuid = uuid
            self.refresh()

    @property
    def base_url(self):
        return "foundation-model/"

    @property
    def detail_url(self):
        return "foundation-model/{uuid}/".format(uuid=self.uuid)

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def knowledgepack_description(self):
        return self._knowledgepack_description

    @knowledgepack_description.setter
    def knowledgepack_description(self, value):
        self._knowledgepack_description = value


class FoundationModelSet(BaseSet):
    def __init__(self, connection, initialize_set=True):
        """Initialize a custom transform set object.

        Args:
            connection
        """
        self._connection = connection
        self._set = None
        self._objclass = FoundationModel
        self._attr_key = "uuid"

        if initialize_set:
            self.refresh()

    @property
    def foundation_models(self):
        return self.objs

    @property
    def get_set_url(self):
        return "foundation-model/"

    def _new_obj_from_dict(self, data):
        """Creates a new object from the response data from the server.

        Args:
            data (dict): contains properties of the object

        Returns:
            obj of type _objclass

        """
        obj = self._objclass(self._connection)
        obj.initialize_from_dict(data)
        return obj



import json
from numpy import float64, int64
import sensiml.base.utility as utility
import logging
import time
from sensiml.datamanager.base import Base, BaseSet
from sensiml.datamanager.labelvalue import LabelValue


logger = logging.getLogger(__name__)


class MetadataValue(LabelValue):
    """Base class for a label object."""

    def __init__(self, connection, project, metadata):
        """Initialize a metadata object.

        Args:
            connection
            project
            label
        """

        if not metadata._metadata:
            raise Exception("Must be metadta not label")

        self._uuid = ""
        self._value = ""
        self._last_modified = ""
        self._created_at = ""
        self._connection = connection
        self._project = project
        self._label = metadata


class MetadataValueSet(BaseSet):
    def __init__(self, connection, project, metadata, initialize_set=True):
        """Initialize a metadata object.

        Args:
            connection
            project
        """
        self._connection = connection
        self._project = project
        self._metadata = metadata
        self._set = None
        self._objclass = MetadataValue
        self._attr_key = "value"

        if initialize_set:
            self.refresh()

    @property
    def metadata_values(self):
        return self.objs

    @property
    def get_set_url(self):
        return "project/{0}/{1}/{2}/labelvalue/".format(
            self._project.uuid, self._metadata._label_or_metadata, self._metadata.uuid
        )

    def _new_obj_from_dict(self, data):
        """Creates a new label from data in the dictionary.

        Args:
            data (dict): contains label_value properties value, uuid

        Returns:
            label object

        """

        obj = self._objclass(self._connection, self._project, self._metadata)
        obj.initialize_from_dict(data)
        return obj

    def __str__(self):
        s = ""
        if self._set:
            metadata = self._set[0]._label
            s = "METADATA\n"
            s += "\tname: " + str(metadata.name) + " uuid: " + str(metadata.uuid) + "\n"
            s += "METADATA VALUES\n"
            for mdv in self._set:
                s += "\tvalue: " + str(mdv.value) + " uuid:" + str(mdv.uuid) + "\n"

        return s



import logging
import time
from requests.exceptions import *
import datetime
import sensiml.base.utility as utility
import pandas as pd
from pandas import DataFrame
import os.path

logger = logging.getLogger(__name__)


class CaptureConfiguration(object):
    """Base class for a Capture."""

    def __init__(
        self,
        connection,
        project,
        name="",
        uuid="",
        configuration=None,
        created_at=None,
        **kwargs
    ):
        """Initialize a Capture instance."""
        self._connection = connection
        self._project = project
        self._uuid = uuid
        self._configuration = configuration
        self._name = name
        self._created_at = created_at

    @property
    def uuid(self):
        """Auto generated unique identifier for the Capture object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        """Auto generated unique identifier for the Capture object"""
        self._uuid = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def created_at(self):
        """Date of the Pipeline creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        if value is None:
            return
        self._created_at = datetime.datetime.strptime(
            value[:-6], "%Y-%m-%dT%H:%M:%S.%f"
        )

    @property
    def configuration(self):
        """The local or server path to the Capture file data"""
        return self._configuration

    @configuration.setter
    def configuration(self, value):
        self._configuration = value

    def insert(self, asynchronous=False):
        """Calls the REST API to insert a new Capture."""
        url = "project/{0}/captureconfiguration/".format(self._project.uuid)
        data = {"name": self.name, "configuration": self.configuration}
        if self.uuid:
            data["uuid"] = self.uuid
        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._uuid = response_data["uuid"]

    def update(self):
        """Calls the REST API to update the capture."""
        url = "project/{0}/captureconfiguration/{1}/".format(
            self._project.uuid, self._uuid
        )

        data = {"name": self.name}
        if self.configuration is not None:
            data["configuration"] = self.configuration

        response = self._connection.request("put", url, data)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API to delete the capture from the server."""
        url = "project/{0}/captureconfiguration/{1}/".format(
            self._project.uuid, self._uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and self populates properties from the server."""
        url = "project/{0}/captureconfiguration/{1}/".format(
            self._project.uuid, self._uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.name = response_data["name"]
            self.configuration = response_data["configuration"]

    @classmethod
    def initialize_from_dict(cls, data):
        """Reads a dictionary or properties and populates a single capture.

        Args:
            capture_dict (dict): contains the capture's 'name' property

        Returns:
            capture object
        """
        assert isinstance(data, dict)

        return CaptureConfiguration(**data)



import logging

from sensiml.datamanager.base import BaseSet
from sensiml.datamanager.label import Label

logger = logging.getLogger(__name__)


class Metadata(Label):
    """Base class for a label object."""

    @property
    def _label_or_metadata(self):
        return "metadata"

    @property
    def _metadata(self):
        return True

    @property
    def metadata(self):
        return True


class MetadataSet(BaseSet):
    def __init__(self, connection, project, initialize_set=True):
        """Initialize a metadata object.

        Args:
            connection
            project
        """

        self._connection = connection
        self._project = project
        self._objclass = Metadata
        self._set = None
        self._attr_key = "name"
        self._data = None

        if initialize_set:
            self.refresh()

    @property
    def metadata(self):
        return self.objs

    @property
    def get_set_url(self):
        return "project/{0}/metadata/".format(self._project.uuid)

    def __str__(self):
        s = ""
        for obj in self.objs:
            s += "name: {0} uuid {1}\n".format(obj.name, obj.uuid)

        return s



import json
from sensiml.datamanager.sandbox import Sandbox
import sensiml.base.utility as utility


class SandboxExistsError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Sandboxes:
    """Base class for a collection of sandboxes."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def build_sandbox_list(self):
        """Populates the function_list property from the server."""
        sandbox_list = {}

        sandbox_response = self.get_sandboxes()
        for sandbox in sandbox_response:
            sandbox_list[sandbox.name] = sandbox

        return sandbox_list

    def get_or_create_sandbox(self, name):
        """Calls the REST API and gets the sandbox by name, if it doesn't exist insert a new sandbox

        Args:
            name (str): name of the sandbox

        Returns:
            sandbox object
        """
        sandbox = self.get_sandbox_by_name(name)

        if sandbox is None:
            print("Sandbox {} does not exist, creating a new sandbox.".format(name))
            sandbox = self.new_sandbox()
            sandbox.name = name
            sandbox.insert()

        return sandbox

    def create_sandbox(self, name, steps):
        """Creates a sandbox using the given name and steps

        Args:
            name (str): name of the sandbox
            steps (list[dict]): list of dictionaries specifying the pipeline steps

        Returns:
            sandbox object
        """
        if self.get_sandbox_by_name(name) is not None:
            raise SandboxExistsError("sandbox {0} already exists.".format(name))
        else:
            sandbox = self.new_sandbox()
            sandbox.name = name
            sandbox._steps = steps
            sandbox.insert()
            return sandbox

    def get_sandbox_by_name(self, name):
        """Calls the REST API and gets the sandbox by name

        Args:
            name (str): name of the sandbox

        Returns:
            sandbox object
        """
        sandbox_list = self.get_sandboxes()
        for sandbox in sandbox_list:
            if sandbox.name == name:
                return sandbox
        return None

    def new_sandbox(self, dict={}):
        """Creates a new sandbox but does not insert in on the server

        Args:
            dict (optional[dict]): contains name and steps properties of the sandbox

        Returns:
            sandbox object
        """
        sandbox = Sandbox(self._connection, self._project)
        if len(dict) != 0:
            sandbox.initialize_from_dict(dict)
        return sandbox

    def get_sandboxes(self):
        """Gets all sandboxes from server and turns them into local sandbox objects.

        Returns:
            list[sandbox]
        """
        # Query the server and get the json
        url = "project/{0}/sandbox/".format(self._project.uuid)
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError as e:
            print(response)
        # Populate each sandbox from the server
        sandboxes = []
        if err is False:
            for sandbox_params in response_data:
                sandboxes.append(self.new_sandbox(sandbox_params))
        return sandboxes

    def __str__(self):
        output_string = "Sandboxes:\n"
        for s in self.get_sandboxes():
            output_string += "    {0}\n".format(s.name)

        return output_string



import json

import sensiml.base.utility as utility
from sensiml.datamanager.function import Function
from sensiml.method_calls import (
    AugmentationCall,
    AugmentationCallSet,
    CaptureFileCall,
    ClassifierCall,
    DataFileCall,
    FeatureFileCall,
    FunctionCall,
    GeneratorCall,
    GeneratorCallSet,
    OptimizerCall,
    QueryCall,
    SelectorCall,
    SelectorCallSet,
    TrainAndValidationCall,
    TrainingAlgorithmCall,
    ValidationMethodCall,
)
from six import string_types


class FunctionExistsError(Exception):
    """Base class for a function exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Functions:
    """Base class for a collection of functions"""

    def __init__(self, connection):
        self._connection = connection
        self._function_list = {}
        self.build_function_list()

    @property
    def function_list(self):
        if not self._function_list:
            self.build_function_list()

        return self._function_list

    @function_list.setter
    def function_list(self, value):
        self._function_list = value

    def build_function_list(self):
        """Populates the function_list property from the server."""
        function_list = {}

        function_response = self.get_functions()
        for function in function_response:
            function_list[function.name] = function

        self._function_list = function_list

    def get_function_by_name(self, name):
        """Gets a function from the server by name.

        Args:
            name (str): function name

        Returns:
            function
        """
        func = self.function_list.get(name, None)
        if func is None:
            raise Exception("No function named {name}".format(name=name))

        return func

    def _new_function_from_dict(self, dict):
        """Creates and populates a function from a set of properties.

        Args:
            dict (dict): contains properties of a function

        Returns:
            function
        """
        function = Function(self._connection)
        function.initialize_from_dict(dict)
        return function

    def get_functions(self, function_type=""):
        """Gets all functions as function objects.

        Args:
            function_type (optional[str]): type of function to retrieve

        Returns:
            list of functions
        """
        url = "transform/"
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        functions = []
        for function_params in response_data:
            functions.append(self._new_function_from_dict(function_params))

        if function_type:
            functions = [f for f in functions if f.type == function_type]

        return functions

    def create_query_call(self, name="Query"):
        """Creates a query call.

        Returns:
            QueryCall
        """
        qy = QueryCall(name)
        return qy

    def create_featurefile_call(self, name):
        """Creates a featurefile call.

        Args:
            name: name of the featurefile

        Returns:
            FeatureFileCall
        """
        return FeatureFileCall(name)

    def create_datafile_call(self, name):
        """Creates a datafile call.

        Args:
            name: name of the datafile

        Returns:
            FeatureFileCall
        """
        return DataFileCall(name)

    def create_capturefile_call(self, name):
        """Creates a capturefile call.

        Args:
            name: name of the featurefile

        Returns:
            FeatureFileCall
        """
        return CaptureFileCall(name)

    def create_train_and_validation_call(self, name="TVO"):
        """Creates an empty train and validation call.

        Returns:
            TrainAndValidationCall
        """
        tvo = TrainAndValidationCall()
        tvo.name = name
        return tvo

    def get_functions_by_type(self, function_type="", subtype=""):
        """Gets all functions or functions of a particular function ype or all functions
        of a particular subtype as function objects.

        Args:
            subtype (optional[str]): subtype to retrieve

        Returns:
            list of functions
        """
        # Populate each function from the server

        if function_type:
            functions = [
                f for f in self.function_list.values() if f.type == function_type
            ]

        elif subtype:
            functions = [f for f in self.function_list.values() if f.subtype == subtype]

        else:
            functions = [f for f in self.function_list.values()]

        return functions

    def __str__(self, function_type=""):
        output_string = ""
        all_functions = self.get_functions_by_type(function_type=function_type)
        for s in set([f.type for f in all_functions]):
            output_string += "{0}:\n".format(s)
            for t in [
                tr
                for tr in all_functions
                if tr.type == s
                and (
                    tr.has_c_version
                    or tr.type not in ["Transform", "Segmenter", "Feature Generator"]
                )
            ]:
                if t.subtype:
                    output_string += "    {0} ({1})\n".format(t.name, t.subtype)
                else:
                    output_string += "    {0}\n".format(t.name)
            output_string += "\n"

        return output_string

    def __call__(self, function_type):
        return self.__str__(function_type)

    def __getitem__(self, key):
        return self.get_function_by_name(key).__str__()

    def _initialize_docstring(self, function):
        ic_dict = function.input_contract
        # oc_dict = function.output_contract
        inputs = {
            i["name"]: None for i in ic_dict if not i.get("handle_by_set", False)
        }  # Parse arguments from the input contract

        # Construct a user-friendly docstring using the description and input/output contracts
        docstring = function.description
        docstring += "\n \nInputs\n---------- \n"

        for i in ic_dict:
            if i.get("default", None):
                docstring += "  {0}: {1} - (default: {2})\n".format(
                    i["name"], i["type"], i["default"]
                )
            else:
                docstring += "  {0}: {1} \n".format(i["name"], i["type"])

        # docstring += '\nOutputs\n---------- \n'
        # for i in oc_dict:
        #     docstring += '  {0} \n'.format(i['type'])

        docstring += "\nUsage\n---------- \n"
        docstring += "For DataFrame inputs, provide a string reference to the\nDataFrame output of a previous step in the pipeline.\n"
        docstring += "For Dataframe output, provide a string name that subsequent\noperations can refer to."

        return ic_dict, inputs, docstring

    def _warn_if_not_transcoded(self, function):
        if (
            function.type in ("Transform", "Segmenter", "Feature Generator")
            and not function.has_c_version
        ):
            print(
                "Warning: {0} is not supported by KnowledgePack code generation. \n"
                "It can be used for cloud-based modeling but cannot be downloaded or flashed to a device. \n".format(
                    function.name
                )
            )

    def create_function_call(self, function, name=None):
        """Creates a function call object.

        Args:
            function (Function object or str): the function for which to create a call object or its
            string name

        Returns:
            function call object or None if the function is not found

        Raises:
            error if no call object can be initialized for the function type
        """
        if name:
            print(
                "The use of the second input parameter is no longer necessary and will be deprecated."
            )

        if isinstance(function, string_types):

            function = self.function_list.get(function, None)

            if function is None:
                print("No function found")
                return None

        self._warn_if_not_transcoded(function)

        try:
            return getattr(
                self, "create_{}_call".format(function.type.lower().replace(" ", "_"))
            )(function)

        except Exception as e:
            print(
                'No function call for function type "{}" found.'.format(
                    function.type.lower()
                )
            )

            raise (e)

    def _create_base_call(self, function, CallFunction, ignore_columns=[]):
        """Creates a base subclass on the fly for the specified function.

        Parses the passed function's I/O contracts and builds class properties accordingly. Returns an instance of the
        new class.
        """

        if isinstance(function, string_types):
            function = self.function_list[function]

        self._warn_if_not_transcoded(function)
        ic_dict, inputs, docstring = self._initialize_docstring(function)

        for column in ignore_columns:
            inputs.pop(column, None)

        for contract in ic_dict:
            if "default" in contract.keys():
                inputs[contract["name"]] = contract["default"]

        # Add docstring to inputs structure and create the new class
        inputs["__doc__"] = docstring
        fname = str(function.name)

        Call = type(fname, (CallFunction,), inputs)

        return Call(function.name, function_type=function.type.lower())

    def create_transform_call(self, function, name=None):
        """Creates a FunctionCall for the specified transform.

        Parses the passed function's I/O contracts and builds class properties accordingly. Returns an instance of the
        new class.

            Args:
                function (Function): function with which to create a transform call

            Returns:
                instance of FunctionCall customized for the input function
        """

        ic_dict, inputs, docstring = self._initialize_docstring(function)

        # Add docstring to inputs structure and create the new class
        inputs["__doc__"] = docstring
        fname = str("kb_" + function.name.lower().replace(" ", "_"))

        for contract in ic_dict:
            if "default" in contract.keys():
                inputs[contract["name"]] = contract["default"]

        Call = type(
            fname, (FunctionCall,), dict({"_input_contract": ic_dict}, **inputs)
        )
        return Call(name=function.name, function_type=function.type.lower())

    def create_segmenter_call(self, function, name=None):

        return self.create_transform_call(function, name=name)

    def create_sampler_call(self, function, name=None):

        return self.create_transform_call(function, name=name)

    def create_feature_generator_call(self, function):
        """Creates a GeneratorCall for the specified generator.

        Args:
            function (Function): function with which to create a feature generator call

        Returns:
            instance of GeneratorCall customized for the input function
        """
        ignore_columns = ["group_columns", "input_data"]

        return self._create_base_call(
            function, GeneratorCall, ignore_columns=ignore_columns
        )

    def create_augmentation_call(self, function):
        """
        Creates a AugmentationCall for the specified time series augmentation.

        Args:
            function (Function): function with which to create a augmentation call

        Returns:
            instance of AugmentationCall customized for the input function

        """
        ignore_columns = [
            "ignore_columns",
            "input_data",
            "label_column",
            "num_features",
        ]

        return self._create_base_call(
            function, AugmentationCall, ignore_columns=ignore_columns
        )

    def create_feature_selector_call(self, function):
        """Creates a SelectorCall for the specified selector.

        Args:
            function (Function): function with which to create a feature selector call

        Returns:
            instance of SelectorCall customized for the input function
        """
        ignore_columns = [
            "ignore_columns",
            "input_data",
            "label_column",
            "num_features",
        ]

        return self._create_base_call(
            function, SelectorCall, ignore_columns=ignore_columns
        )

    def create_validation_method_call(self, function, name=None):
        """Creates a ValidationMethodCall for the specified function.

        Args:
            function (Function): function with which to create a validation method call

        Returns:
            instance of ValidationMethodCall customized for the input function
        """
        return self._create_base_call(function, ValidationMethodCall)

    def create_classifier_call(self, function):
        """Creates a ClassifierCall for the specified function.

        Args:
            function (Function): function with which to create a classifier call

        Returns:
            instance of ClassifierCall customized for the input function
        """
        return self._create_base_call(function, ClassifierCall)

    def create_optimizer_call(self, function, name=None):
        """Creates an OptimizerCall for the specified function.

        Args:
            function (Function): function with which to create an optimizer call

        Returns:
            instance of OptimizerCall customized for the input function
        """
        return self._create_base_call(function, OptimizerCall)

    def create_generator_subtype_call(self, subtype):
        """Creates a GeneratorCall for the specified subtype of generators.

        Uses the input contract of one function of the subtype (assumed to be constant for all generators of the subtype),
        and builds class properties accordingly.

            Args:
                subtype (str): name of the subtype for which to create a generator call

            Returns:
                instance of GeneratorCall customized for the subtype
        """

        function_list = self.get_functions_by_type(subtype=subtype)
        if len(function_list) > 0:
            function = function_list[0]
            ic_dict, inputs, docstring = self._initialize_docstring(function)
        else:
            print(
                "Error: No generators of subtype {0}; length of function list {1}".format(
                    subtype, len(function_list)
                )
            )
            return None

        for function in function_list:
            self._warn_if_not_transcoded(function)

        ignore_columns = ["group_columns", "input_data"]

        for column in ignore_columns:
            inputs.pop(column, None)

        for contract in ic_dict:
            if "default" in contract.keys():
                inputs[contract["name"]] = contract["default"]

        # Add docstring to inputs structure and create the new class
        inputs["__doc__"] = docstring
        inputs["_subtype"] = subtype

        Call = type(subtype, (GeneratorCall,), inputs)
        return Call(function.name, "generator")

    def create_generator_call_set(self, name="GeneratorCallSet"):
        """Creates an empty generator call set."""
        gcs = GeneratorCallSet(name)
        return gcs

    def create_selector_call_set(self, name="SelectorCallSet"):
        """Creates an empty selector call set."""
        scs = SelectorCallSet(name)
        return scs

    def create_augmentation_call_set(self, name="AugmentationCallSet"):
        """Creates an empty augmentation call set."""
        acs = AugmentationCallSet(name)
        return acs

    def create_training_algorithm_call(self, function, name=None):
        """Creates a training algorithm method call for the specified function.

        Args:
            function (Function): function with which to create a training algorithm call

        Returns:
            instance of TrainingAlgorithmCall customized for the input function
        """
        return self._create_base_call(function, TrainingAlgorithmCall)



class empty:
    pass


param_tabs = ",\n" + " " * 4 * 8


def function_help(func):
    filter_list = [
        "group_columns",
        "label_column",
        "ignore_columns",
        "passthrough_columns",
        "classifiers",
        "validation_methods",
    ]
    params = []
    for item in func.input_contract:
        key = item.get("name")
        value = item.get("default", None)
        key_type = item.get("type", None)
        options = item.get("options", None)
        if value is not None:
            if key_type == "str":
                params.append('"{}": "{}"'.format(key, value))
            else:
                params.append('"{}": {}'.format(key, value))

            if options is not None and key not in filter_list:
                values = map(lambda x: x["name"], options)
                params[-1] += ", #options: " + "<{}>".format("/".join(values))
        elif options and key not in filter_list:
            values = map(lambda x: x["name"], options)
            params.append('"{}": <{}>'.format(key, "/".join(values)))
        elif (
            key_type in ["int", "float", "str", "numeric", "list", "dict"]
            and key not in filter_list
        ):
            params.append('"{}": {}'.format(key, "<{}>".format(key_type)))

    params.append("")

    if func.type == "Feature Generator":
        return """client.pipeline.add_feature_generator([{{'name':'{}', 'params':{{{}}}}}])""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Feature Selector":
        return """client.pipeline.add_feature_selector([{{'name':'{}', 'params':{{{}}}}}], params={{'number_of_features':(int)}})""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Augmentation":
        return """client.pipeline.add_augmentation(["name":"{}", "params":{{{}}}])""".format(
            func.name, param_tabs.join(params)
        )

    if func.type in ["Segmenter", "Transform", "Sampler"]:
        return """client.pipeline.add_transform("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Training Algorithm":
        return """client.pipeline.set_training_algorithm("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Validation Method":
        return """client.pipeline.set_validation_method("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Classifier":
        return """client.pipeline.set_classifier("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    return None


class hold_function:
    def __init__(self, func):
        self.func = func

    def create_function(self):
        try:
            docs = function_help(self.func)
        except:
            docs = "unable to generate this function."
        try:
            get_ipython().set_next_input(docs, replace=True)
        except:
            print(docs)

        return docs


class Snippets:
    def __init__(self, function_df, function_list):
        self._function_df = function_df
        self._function_list = function_list
        self._attach_types()

    def _attach_types(self):
        type_list = self._function_df.TYPE.unique()
        for func_type in type_list:
            if not func_type:
                continue

            if func_type == "Transform":
                for subtype in self._function_df[
                    self._function_df.TYPE == func_type
                ].SUBTYPE.unique():
                    setattr(
                        self,
                        subtype.replace(" ", "_")
                        if "Filter" in subtype
                        else subtype.replace(" ", "_") + "_Transform",
                        self._attach_functions(subtype=subtype),
                    )
            else:
                setattr(
                    self,
                    func_type.replace(" ", "_"),
                    self._attach_functions(func_type=func_type),
                )
        # setattr(self, 'Auto', self._attach_seeds())

    def _attach_functions(self, func_type=None, subtype=None):
        cls = empty()
        if subtype:
            funcs = self._function_df[self._function_df.SUBTYPE == subtype]
        else:
            funcs = self._function_df[self._function_df.TYPE == func_type]
        for func in funcs.NAME:
            # functions names starting with numbers not allowed, check for them here
            sfunc = func
            if sfunc[:1] in "0123456789":
                sfunc = "FG_" + func
            setattr(
                cls,
                sfunc.replace(" ", "_").replace("-", ""),
                hold_function(self._function_list[func]).create_function,
            )

        return cls


def get_group_columns(steps):
    for step in steps:
        inputs = step.get("inputs", None)
        if inputs:
            return inputs.get("group_columns")


def get_label_column(steps):
    for step in steps:
        if step.get("type", None) == "tvo":
            return step.get("label_column", None)


def pipeline_function_help(
    function_list,
    step,
    filter_list=[
        "group_columns",
        "label_column",
        "ignore_columns",
        "passthrough_columns",
        "classifiers",
        "validation_methods",
        "input_data",
        "return_segment_index",
    ],
):

    function_name = step.get("name", None)
    if function_name is None:
        function_name = step.get("function_name", None)
    if function_name is None:
        return None

    func = function_list[function_name]

    params = []
    for item in func.input_contract:
        key = item.get("name")
        if key not in filter_list:
            value = step["inputs"].get(key, None)
            if value is not None:
                if item.get("type", None) == "str":
                    params.append('"{}":"{}"'.format(key, value))
                else:
                    params.append('"{}":{}'.format(key, value))

    if func.type == "Feature Generator":
        return """{{'name':'{}', 'params':{{{}}}}}""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Feature Selector":
        return """{{'name':'{}', 'params':{{{}}}}}, """.format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Augmentation":
        return """\n{}{{"name": "{}", "params": {{{}}}}},""".format(
            " " * 32, func.name, param_tabs.join(params)
        )

    if func.type in ["Segmenter", "Transform", "Sampler"]:
        return """client.pipeline.add_transform("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Training Algorithm":
        return """client.pipeline.set_training_algorithm("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Validation Method":
        return """client.pipeline.set_validation_method("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    if func.type == "Classifier":
        return """client.pipeline.set_classifier("{}", params={{{}}})""".format(
            func.name, param_tabs.join(params)
        )

    return None


def build_pipeline(function_list, steps):
    pipeline_steps = ["client.pipeline.reset()"]

    for step in steps:
        if step.get("type") == "generatorset":
            feature_generators = []
            for generator in step["set"]:
                feature_generators.append(
                    pipeline_function_help(function_list, generator)
                )
            pipeline_steps.append(
                """client.pipeline.add_feature_generator([{}])""".format(
                    ",\n{}".format(" " * 40).join(feature_generators)
                )
            )

        elif step.get("type") == "selectorset":
            feature_selectors = []
            for selector in step["set"]:
                feature_selectors.append(
                    pipeline_function_help(function_list, selector)
                )
            pipeline_steps.append(
                "client.pipeline.add_feature_selector([{}])"
                "".format("\n{}".format(" " * 40).join(feature_selectors))
            )

        elif step.get("type") == "augmentationset":
            data_augmentors = []
            for augmentor in step["set"]:
                data_augmentors.append(pipeline_function_help(function_list, augmentor))
            pipeline_steps.append(
                "client.pipeline.add_augmentation([{}])"
                "".format("{}".format(" " * 40).join(data_augmentors))
            )

        elif step.get("type") == "tvo":
            for key in ["classifiers", "validation_methods", "optimizers"]:
                if step[key]:
                    pipeline_steps.append(
                        pipeline_function_help(function_list, step[key][0])
                    )

            pipeline_steps.append(
                "client.pipeline.set_tvo({{'validation_seed':{}}})".format(
                    step.get("validation_seed", 0)
                )
            )

        elif step.get("type", None) == "featurefile":
            pipeline_steps.append(
                """client.pipeline.set_input_data("{}", data_columns={}, \n{}group_columns={},\n{}label_column="{}")""".format(
                    step.get("name", None),
                    step.get("data_columns", None),
                    " " * 40,
                    get_group_columns(steps),
                    " " * 40,
                    get_label_column(steps),
                )
            )

        elif step.get("type", None) == "query":
            pipeline_steps.append(
                """client.pipeline.set_input_query("{query_name}", use_session_preprocessor={use_session_preprocessor})""".format(
                    query_name=step.get("name", None),
                    use_session_preprocessor=step.get("use_session_preprocessor", True),
                )
            )
        else:
            pipeline_steps.append(pipeline_function_help(function_list, step))

    return pipeline_steps


def generate_pipeline(function_list, steps, replace=True):
    pipeline_steps = build_pipeline(function_list, steps)

    if replace:
        pipeline_string = "\n\n".join([x for x in pipeline_steps if x])
        get_ipython().set_next_input(pipeline_string, replace=True)

    return pipeline_steps



import json
import logging
import time

import sensiml.base.utility as utility
from numpy import float64, int64
from sensiml.datamanager.base import Base, BaseSet

logger = logging.getLogger(__name__)


class LabelType(object):
    Int = "integer"
    Float = "float"
    String = "string"


class Label(Base):
    """Base class for a label object."""

    _fields = [
        "uuid",
        "name",
        "value_type",
        "is_dropdown",
        "metadata",
        "last_modified",
        "created_at",
    ]

    _field_map = {"value_type": "type"}

    def __init__(self, connection, project):
        """Initialize a metadata object.

        Args:
            connection
            project
        """
        self._uuid = ""
        self._name = ""
        self._value_type = "string"
        self._is_dropdown = False
        self._last_modified = ""
        self._created_at = ""
        self._connection = connection
        self._project = project

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """The name property of the metadata object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def value_type(self):
        """The data type of the metadata object"""
        return self._value_type

    @value_type.setter
    def value_type(self, value):
        if value not in ["string", "int", "float"]:
            raise Exception("Invalid value type")
        self._value_type = value

    @property
    def _label_or_metadata(self):
        return "label"

    @property
    def _metadata(self):
        return False

    @property
    def metadata(self):
        return False

    @property
    def is_dropdown(self):
        return self._is_dropdown

    @is_dropdown.setter
    def is_dropdown(self, value):
        if not isinstance(value, bool):
            raise Exception("Invalid Value")

        self._is_dropdown = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/{1}/".format(
            self._project.uuid,
            self._label_or_metadata,
        )
        data = self._to_representation()
        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

        return response

    def update(self):
        """Calls the REST API and updates the object on the server."""
        url = "project/{0}/{1}/{2}/".format(
            self._project.uuid, self._label_or_metadata, self.uuid
        )

        data = self._to_representation()
        response = self._connection.request("put", url, data)
        response_data, err = utility.check_server_response(response)

        return response

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/{1}/{2}/".format(
            self._project.uuid, self._label_or_metadata, self.uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

        return response

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/{1}/{2}/".format(
            self._project.uuid, self._label_or_metadata, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single metadata object.

        Args:
            data (dict): contains the uuid, name, type
        """
        self.uuid = data["uuid"]
        self.name = data["name"]
        if data["type"] == LabelType.Int:
            self.value_type = int(data["type"])
        elif data["type"] == LabelType.Float:
            self.value_type = float(data["type"])
        else:
            self.value_type = data["type"]

        self._data = data


class LabelSet(BaseSet):
    def __init__(self, connection, project, initialize_set=True):
        """Initialize a metadata object.

        Args:
            connection
            project
        """
        self._connection = connection
        self._project = project
        self._set = None
        self._objclass = Label
        self._attr_key = "name"

        if initialize_set:
            self.refresh()

    @property
    def labels(self):
        return self.objs

    @property
    def get_set_url(self):
        return "project/{0}/label/".format(self._project.uuid)
